/*
 * Copyright (c) 2001 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*	CFFileUtilities.c
	Copyright 1999-2001, Apple, Inc. All rights reserved.
	Responsibility: Christopher Kane
*/

#include "CFFileUtilities.h"
#include "CFInternal.h"
#include "CFPriv.h"
#if defined(__WIN32__)
    #include <io.h>
    #include <fcntl.h>
    #include <sys/stat.h>
    #include <errno.h>
    #define timeval xxx_timeval
    #define BOOLEAN xxx_BOOLEAN
        #include <windows.h>
    #undef BOOLEAN
    #undef timeval
#elif defined(__MACOS8__)
    #include <string.h>
    #include <unistd.h>
    #include <stat.h>
    #include <types.h>
    #include <Errors.h>
    #include <Files.h>
    #include <Script.h>
    #include <Gestalt.h>
#else
    #include <string.h>
    #include <unistd.h>
    #include <dirent.h>
    #include <sys/stat.h>
    #include <sys/types.h>
    #include <pwd.h>
    #include <fcntl.h>
    #include <errno.h>
    #include <stdio.h>
#endif

#if defined(__MACOS8__)
#if !defined(ENOENT)
#define ENOENT fnfErr
#endif
#endif

#if defined(__WIN32__)
    #define CF_OPENFLGS	(_O_BINARY|_O_NOINHERIT)
#else
    #define CF_OPENFLGS	(0)
#endif

#if defined(__MACOS8__)
#include <CodeFragments.h>
#include <MacErrors.h>

static Ptr __CFGropeAroundForMacOS8Symbol(ConstStr255Param name) {
     static const char *libraries[] = {"\pCarbonLib", "\pInterfaceLib", "\pPrivateInterfaceLib"};
     int idx;
     for (idx = 0; idx < sizeof(libraries) / sizeof(libraries[0]); idx++) {
         CFragConnectionID connID;  /* We get the connections ONLY if already available. */
         OSErr err = GetSharedLibrary(libraries[idx], kPowerPCCFragArch, kFindCFrag, &connID, NULL, NULL);
         if (err == noErr) {
             Ptr cfmfunc = NULL;
             CFragSymbolClass symbolClass;
             err = FindSymbol(connID, name, &cfmfunc, &symbolClass);
             if (err == noErr && symbolClass == kTVectorCFragSymbol) {
                 return cfmfunc;
	     }
         }
    }
    return NULL;
}

static Boolean __CFMacOS8HasFSRefs() {
       static Boolean sHasFSRefs = (Boolean) -1;
       if ( sHasFSRefs == (Boolean) -1 ) {
               long result;
               sHasFSRefs = Gestalt( gestaltFSAttr, &result ) == noErr &&
                                   ( result & (1 << gestaltHasHFSPlusAPIs) ) != 0;
       }
       return sHasFSRefs;
}

static OSErr __CF_FSGetCatalogInfo_internal(const FSRef *A, FSCatalogInfoBitmap B, FSCatalogInfo *C, HFSUniStr255 *D, FSSpec *E, FSRef *F) {
    static OSErr (*cfmfunc)(const FSRef *, FSCatalogInfoBitmap, FSCatalogInfo *, HFSUniStr255 *, FSSpec *, FSRef *) = NULL;
    static Boolean looked = false;
    if (!looked && __CFMacOS8HasFSRefs()) {
        cfmfunc = __CFGropeAroundForMacOS8Symbol("\pFSGetCatalogInfo");
        looked = true;
    }
    return (cfmfunc != NULL) ? cfmfunc(A, B, C, D, E, F) : cfragNoSymbolErr;
}

#define FSGetCatalogInfo(A, B, C, D, E, F) __CF_FSGetCatalogInfo_internal(A, B, C, D, E, F)

Boolean _CFGetFSSpecFromPathString(CFAllocatorRef alloc, CFStringRef filePath, void *voidspec) {
    // This should be rewritten to just explode the path and go component by component.  REW, 9/16/99
    FSSpec *spec = (FSSpec *)voidspec;
    Boolean result = true;
    SInt32 pathLen = CFStringGetLength(filePath);
    Str255 pStr;
    CFStringRef colonStr = CFSTR(":");

    CFRange strRange = CFRangeMake(0, pathLen);
    CFRange foundRange;
    CFStringRef subStr;

    int16_t volRefNum = 0;
    SInt32 dirID = 0;

    CInfoPBRec catInfoRec;
    OSErr err;

    if (strRange.length > 250) {
        // Use FSMakeFSSpec to find the volRefNum from the first path component.
        // Find the first colon and grab the disk name in a pascal string.
        if (!CFStringFindWithOptions(filePath, colonStr, strRange, 0, &foundRange)) {
            // Can't find a colon.  That's bad.
            result = false;
        } else {
            subStr = CFStringCreateWithSubstring(alloc, filePath, CFRangeMake(strRange.location, foundRange.location - strRange.location));
            if (!CFStringGetPascalString(subStr, pStr, 255, CFStringFileSystemEncoding())) {
		if (!CFStringGetPascalString(subStr, pStr, 255, kCFStringEncodingMacRoman)) {
                    result = false;
		}
            }
            CFRelease(subStr);
            strRange = CFRangeMake(foundRange.location, pathLen - foundRange.location);
            // Use FSMakeFSSpec on the disk name.
            if (result) {
                err = FSMakeFSSpec(volRefNum, dirID, pStr, spec);
                if (err != noErr && err != fnfErr) { // fnfErr means the file does not exist, but the fsspec is still valid
                    result = false;
                } else {
                    volRefNum = spec->vRefNum;
                    dirID = spec->parID;
                }
            }
        }
    }
    // Now we have a volRefNum and the dirID of the root directory of that vol and strRange points to all the path components beyond the disk name.

    if (result) {
        while (strRange.length > 250) {
            // Carve off a chunk of path components...
            if (!CFStringFindWithOptions(filePath, colonStr, CFRangeMake(strRange.location, 250), kCFCompareBackwards, &foundRange)) {
                // Can't find a colon.  That's bad.
                result = false;
                break;
            }
            subStr = CFStringCreateWithSubstring(alloc, filePath, CFRangeMake(strRange.location, foundRange.location - strRange.location));
            if (!CFStringGetPascalString(subStr, pStr, 255, CFStringFileSystemEncoding())) {
		if (!CFStringGetPascalString(subStr, pStr, 255, kCFStringEncodingMacRoman)) {
                    result = false;
		}
            }
            CFRelease(subStr);
            if (!result) {
                break;
            }
            strRange = CFRangeMake(foundRange.location, pathLen - foundRange.location);

            // pStr is now the partial path we want to process and strRange contains the range of the rest of the string yet to be dealt with.
            // Use PBGetCatInfo to find the dirID of the last directory in the partial path.
            catInfoRec.dirInfo.ioCompletion = NULL;
            catInfoRec.dirInfo.ioNamePtr = pStr;
            catInfoRec.dirInfo.ioVRefNum = volRefNum;
            catInfoRec.dirInfo.ioFDirIndex = 0;
            catInfoRec.dirInfo.ioDrDirID = dirID;
            err = PBGetCatInfoSync(&catInfoRec);
            if (err != noErr) {
                result = false;
                break;
            }
            dirID = catInfoRec.dirInfo.ioDrDirID;
        }
    }

    if (result) {
        /* strRange should now point to the part of the path that should be sent to FSMakeFSSpec.  volRefNum and dirID should be correct for the strRange substring of filePath to be evaluated as a relative path. */
        subStr = CFStringCreateWithSubstring(alloc, filePath, strRange);
        if (!CFStringGetPascalString(subStr, pStr, 255, CFStringFileSystemEncoding())) {
	    if (!CFStringGetPascalString(subStr, pStr, 255, kCFStringEncodingMacRoman)) {
                result = false;
	    }
        }
        CFRelease(subStr);
    }

    if (result) {
        err = FSMakeFSSpec(volRefNum, dirID, pStr, spec);
        if (err != noErr && err != fnfErr) {  // fnfErr means the file doesn't exist, but the FSSpec is still valid
            result = false;
        } 
    }

    return result;
}

CFStringRef _CFCreateStringWithHFSPathFromFSSpec(CFAllocatorRef alloc, const void *voidspec) {
    CFStringRef result = NULL;
    CFStringRef str = NULL;
    Boolean error = false;
    FSSpec *spec = (FSSpec *)voidspec;
    CFMutableArrayRef components;
    
    CInfoPBRec catInfoRec;
    Str255 dirName;
    OSErr oserr;

    if (spec->parID == fsRtParID) {
        return CFRetain(CFSTR("/"));
    } else if (spec->parID == fsRtDirID) {
        CFStringRef result;
        str = CFStringCreateWithPascalString(alloc, spec->name, CFStringFileSystemEncoding());
	if (!str) {
            str = CFStringCreateWithPascalString(alloc, spec->name, kCFStringEncodingMacRoman);
	}
        if (str) {
            result = CFStringCreateWithFormat(alloc, NULL, CFSTR("/:%@"), str);
            CFRelease(str);
            return result;
        } else {
            return NULL;
        }
    }

    components = CFArrayCreateMutable(alloc, 0, &kCFTypeArrayCallBacks);
    str = CFStringCreateWithPascalString(alloc, spec->name, CFStringFileSystemEncoding());
    if (!str) {
        str = CFStringCreateWithPascalString(alloc, spec->name, kCFStringEncodingMacRoman);
    }
    if (str) {
        CFArrayInsertValueAtIndex(components, 0, str);
        CFRelease(str);
    } else {
        CFRelease(components);
        return NULL;
    }

    catInfoRec.dirInfo.ioNamePtr = dirName;
    catInfoRec.dirInfo.ioVRefNum = spec->vRefNum;
    catInfoRec.dirInfo.ioDrParID = spec->parID;
    catInfoRec.dirInfo.ioFDirIndex = -1;

    while (1) {
        catInfoRec.dirInfo.ioDrDirID = catInfoRec.dirInfo.ioDrParID;
        oserr = PBGetCatInfoSync(&catInfoRec);
        if (oserr != noErr) {
            error = false;
            break;
        }
        str = CFStringCreateWithPascalString(alloc, catInfoRec.dirInfo.ioNamePtr, CFStringFileSystemEncoding());
	if (!str) {
            str = CFStringCreateWithPascalString(alloc, catInfoRec.dirInfo.ioNamePtr, kCFStringEncodingMacRoman);
	}
        if (str) {
            CFArrayInsertValueAtIndex(components, 0, str);
            CFRelease(str);
        } else {
            CFRelease(components);
            return NULL;
        }
        if (catInfoRec.dirInfo.ioDrDirID == fsRtDirID) {
            break;
        }
    }
    if (!error) {
        result = CFStringCreateByCombiningStrings(alloc, components, CFSTR(":"));
    }
    CFRelease(components);
    
    return result;
}

CFStringRef _CFCreateStringWithHFSPathFromFSRef(CFAllocatorRef allocator, const void *voidRef) {
    FSRef fsRef, fsRef2;
    FSRef *parRef, *fileRef;
    CFStringRef result = NULL;
    CFMutableStringRef mString, extString;
    OSErr err = noErr;

    if (!__CFMacOS8HasFSRefs()) {
	return NULL;
    }
    mString = CFStringCreateMutable(allocator, 0);  // The path
    extString = CFStringCreateMutableWithExternalCharactersNoCopy(allocator, NULL, 0, 0, kCFAllocatorNull);
		// To hold the directory names as we retrieve them.

    fileRef = (FSRef *)voidRef;
    parRef = &fsRef2;
    while (err == noErr) {
        HFSUniStr255 name;
        FSCatalogInfo catInfo;
        err = FSGetCatalogInfo(fileRef, kFSCatInfoParentDirID, &catInfo, &name, NULL, parRef);
        if (err == noErr) {
            CFStringSetExternalCharactersNoCopy(extString, name.unicode, name.length, 255);
            CFStringInsert(mString, 0, extString);
            if (catInfo.parentDirID == fsRtParID) {
                break;
            } else {
                CFStringInsert(mString, 0, CFSTR(":"));
            }
            fileRef = parRef;
            parRef = (fileRef == &fsRef) ? &fsRef2 : &fsRef;
        }
    }
    CFRelease(extString);
    if (err == noErr) {
        result = mString;
    } else {
        CFRelease(mString);
        result = NULL;
    }
    return result;
}
#endif

__private_extern__ CFStringRef _CFCopyExtensionForAbstractType(CFStringRef abstractType) {
    return (abstractType ? CFRetain(abstractType) : NULL);
}


__private_extern__ Boolean _CFCreateDirectory(const char *path) {
#if defined(__WIN32__)
    return CreateDirectoryA(path, (LPSECURITY_ATTRIBUTES)NULL);
#elif defined(__MACOS8__)
	CFStringRef pathStr = CFStringCreateWithCStringNoCopy(NULL, path, CFStringFileSystemEncoding(), kCFAllocatorNull);
	FSSpec spec;
	Boolean result = false;
	if (pathStr && _CFGetFSSpecFromPathString(NULL, pathStr, (void *)&spec) && FSpDirCreate(&spec, smSystemScript, NULL) == noErr) {
		result = true;
	}
	if (pathStr) CFRelease(pathStr);
	return result;
#else
    return ((mkdir(path, 0777) == 0) ? true : false);
#endif
}

__private_extern__ Boolean _CFRemoveDirectory(const char *path) {
#if defined(__WIN32__)
    return RemoveDirectoryA(path);
#elif defined(__MACOS8__)
	CFStringRef pathStr = CFStringCreateWithCStringNoCopy(NULL, path, CFStringFileSystemEncoding(), kCFAllocatorNull);
	FSSpec spec;
	Boolean result = false;
	if (pathStr && _CFGetFSSpecFromPathString(NULL, pathStr, (void *)&spec) && FSpDelete(&spec) == noErr) {
		result = true;
	}
	if (pathStr) CFRelease(pathStr);
	return result;
#else
    return ((rmdir(path) == 0) ? true : false);
#endif
}

__private_extern__ Boolean _CFDeleteFile(const char *path) {
#if defined(__MACOS8__)
#if 0
    FSRef fsRef;
    Boolean isDir;
    if (FSPathMakeRef(path, &fsRef, &isDir) != noErr || isDir != false) {
        return false;
    } else {
        return (FSDeleteObject(&fsRef) == noErr);
    }
#else 0
	CFStringRef pathStr = CFStringCreateWithCStringNoCopy(NULL, path, CFStringFileSystemEncoding(), kCFAllocatorNull);
	FSSpec spec;
	Boolean result = false;
	if (pathStr && _CFGetFSSpecFromPathString(NULL, pathStr, (void *)&spec) && FSpDelete(&spec) == noErr) {
		result = true;
	}
	if (pathStr) CFRelease(pathStr);
	return result;
#endif 0
#elif defined(__WIN32__)
    return DeleteFileA(path);
#else
    return unlink(path) == 0;
#endif
}

__private_extern__ Boolean _CFReadBytesFromFile(CFAllocatorRef alloc, CFURLRef url, void **bytes, CFIndex *length, CFIndex maxLength) {
#if !defined(__MACOS8__)
    // maxLength is the number of bytes desired, or 0 if the whole file is desired regardless of length.
    struct stat statBuf;
    int fd = -1;
    char path[CFMaxPathSize];
    if (!CFURLGetFileSystemRepresentation(url, true, path, CFMaxPathSize)) {
        return false;
    }

    *bytes = NULL;
#if defined(__WIN32__)
    fd = open(path, O_RDONLY|CF_OPENFLGS, 0666|_S_IREAD);
#else
    fd = open(path, O_RDONLY|CF_OPENFLGS, 0666);
#endif
    if (fd < 0) {
        return false;
    }
    if (fstat(fd, &statBuf) < 0) {
        int saveerr = thread_errno();
        close(fd);
        thread_set_errno(saveerr);
        return false;
    }
    if ((statBuf.st_mode & S_IFMT) != S_IFREG) {
        close(fd);
        thread_set_errno(EACCES);
        return false;
    }
    if (statBuf.st_size == 0) {
        *bytes = CFAllocatorAllocate(alloc, 4, 0); // don't return constant string -- it's freed!
        *length = 0;
    } else {
        CFIndex desiredLength;
        if ((maxLength >= statBuf.st_size) || (maxLength == 0)) {
            desiredLength = statBuf.st_size;
        } else {
            desiredLength = maxLength;
        }
        *bytes = CFAllocatorAllocate(alloc, desiredLength, 0);
        if (read(fd, *bytes, desiredLength) < 0) {
            CFAllocatorDeallocate(alloc, *bytes);
            close(fd);
            return false;
        }
        *length = desiredLength;
    }
    close(fd);
    return true;
#else // defined(__MACOS8__)
//Boolean _CFReadBytesFromFile(CFAllocatorRef alloc, CFURLRef url, void **bytes, CFIndex *length, CFIndex maxLength) {
    FSSpec fsSpec;
    short refNum;
    long fileLength;
    Boolean success;
    if (!_CFGetFSSpecFromURL(alloc, url, &fsSpec) ||
        FSpOpenDF(&fsSpec, fsRdPerm, &refNum) != noErr) {
        return false;
    }
    if (SetFPos(refNum, fsFromStart, 0) != noErr ||
        GetEOF(refNum, &fileLength) != noErr) {
        success = false;
    }
    
    if (fileLength == 0) {
        *bytes = CFAllocatorAllocate(alloc, 4, 0); // don't return constant string -- it's freed!
        *length = 0;
        success = true;
    } else {
        long desiredLength;
        if ((maxLength >= fileLength) || (maxLength == 0)) {
            desiredLength = fileLength;
        } else {
            desiredLength = maxLength;
        }
        *bytes = CFAllocatorAllocate(alloc, desiredLength, 0);
        if (FSRead(refNum, &desiredLength, *bytes) != noErr) {
            CFAllocatorDeallocate(alloc, *bytes);
            success = false;
        } else {
            *length = desiredLength;
            success = true;
        }
    }
    FSClose(refNum);
    return success;
#endif
}

#if !defined(__MACOS8__)
__private_extern__ Boolean _CFWriteBytesToFile(CFURLRef url, const void *bytes, CFIndex length) {
    struct stat statBuf;
    int fd = -1;
    int mode, mask;
    char path[CFMaxPathSize];
    if (!CFURLGetFileSystemRepresentation(url, true, path, CFMaxPathSize)) {
        return false;
    }

#if defined(__WIN32__)
    mask = 0;
#else
    mask = umask(0);
    umask(mask);
#endif
    mode = 0666 & ~mask;
    if (0 == stat(path, &statBuf)) {
        mode = statBuf.st_mode;
    } else if (thread_errno() != ENOENT) {
        return false;
    }
#if defined(__WIN32__)
    fd = open(path, O_WRONLY|O_CREAT|O_TRUNC|CF_OPENFLGS, 0666|_S_IWRITE);
#else
    fd = open(path, O_WRONLY|O_CREAT|O_TRUNC|CF_OPENFLGS, 0666);
#endif
    if (fd < 0) {
        return false;
    }
    if (length && write(fd, bytes, length) != length) {
        int saveerr = thread_errno();
        close(fd);
        thread_set_errno(saveerr);
        return false;
    }
#if defined(__WIN32__)
    {
        void *nativeHandle = _get_osfhandle(fd);
        FlushFileBuffers(nativeHandle);
    }
#else
    fsync(fd);
#endif
    close(fd);
    return true;
}

#else // __MACOS8__

Boolean _CFWriteBytesToFile(CFURLRef url, const void *bytes, CFIndex length) {
#warning Should we destroy the resource fork as well?  REW, 6/5/2000
    FSSpec fsSpec;
    short refNum;
    OSStatus status;
    if (!_CFGetFSSpecFromURL(CFGetAllocator(url), url, &fsSpec)) {
        return false;
    }
    status = FSpOpenDF(&fsSpec, fsWrPerm, &refNum);
    if (status != noErr && status != fnfErr) {
        return false;
    } else if (status == fnfErr) {
        // Nitin sez - Pass 0x0 for the type and creator. This mimics the catalog info of files that are created by, for example, touch, which is probably what you want to do.  6/5/2000
        if (FSpCreate(&fsSpec, 0x0, 0x0, smSystemScript) != noErr) {
            return false;
        } else if (FSpOpenDF(&fsSpec, fsWrPerm, &refNum) != noErr) {
            return false;
        }
    }
    
    if (SetFPos(refNum, fsFromStart, 0) != noErr ||
        FSWrite(refNum, &length, bytes) != noErr ||
        SetEOF(refNum, length) != noErr) {
        FSClose(refNum);
        return false;
    }
    FSClose(refNum);
    return true;
}
#endif


/* On Mac OS 8/9, one of dirSpec and dirURL must be non-NULL.  On all other platforms, one of path and dirURL must be non-NULL
If both are present, they are assumed to be in-synch; that is, they both refer to the same directory.  */
__private_extern__ CFMutableArrayRef _CFContentsOfDirectory(CFAllocatorRef alloc, char *dirPath, void *dirSpec, CFURLRef dirURL, CFStringRef matchingAbstractType) {
    CFMutableArrayRef files = NULL;
    Boolean releaseBase = false;
    CFIndex pathLength = dirPath ? strlen(dirPath) : 0;
    // MF:!!! Need to use four-letter type codes where appropriate.
    CFStringRef extension = (matchingAbstractType ? _CFCopyExtensionForAbstractType(matchingAbstractType) : NULL);
    CFIndex extLen = (extension ? CFStringGetLength(extension) : 0);
    uint8_t extBuff[CFMaxPathSize];

#if defined(__WIN32__)
    /* Windows Variables */
    /* The Win32 code has not been updated for:
        path has been renamed dirPath
        base has been renamed dirURL
        dirPath may be NULL (in which case dirURL is not)
        if dirPath is NULL, pathLength is 0
    */
    WIN32_FIND_DATA file;
    HANDLE handle;
#elif defined(__svr4__) || defined(__hpux__) || defined(__LINUX__) || defined(__FREEBSD__)
    /* Solaris and HPUX Variables */
    /* The Solaris and HPUX code has not been updated for:
        base has been renamed dirURL
        dirPath may be NULL (in which case dirURL is not)
        if dirPath is NULL, pathLength is 0
    */
    DIR *dirp;
    struct dirent *dp;
    int err;
#elif defined(__MACH__)
    /* Mac OS X Variables */
    int fd, numread;
    long basep;
    char dirge[8192];
    uint8_t pathBuf[CFMaxPathSize];
#elif defined(__MACOS8__)
    /* Mac OS 8 Variables */
    FSSpec specBuf;
#endif

    
    if (extLen > 0) {
        CFStringGetBytes(extension, CFRangeMake(0, extLen), CFStringFileSystemEncoding(), 0, false, extBuff, CFMaxPathSize, &extLen);
        extBuff[extLen] = '\0';
    }
    
#if defined(__WIN32__)
    /* Windows Implementation */
    
    if (pathLength + 2 >= CFMaxPathLength) {
        if (extension) {
            CFRelease(extension);
        }
        return NULL;
    }
    path[pathLength] = '\'';
    path[pathLength + 1] = '*';
    path[pathLength + 2] = '\0';
    handle = FindFirstFileA(path, &file);
    if (INVALID_HANDLE_VALUE == handle) {
        path[pathLength] = '\0';
        if (extension) {
            CFRelease(extension);
        }
        return NULL;
    }
    files = CFArrayCreateMutable(alloc, 0, &kCFTypeArrayCallBacks);

    do {
        CFURLRef fileURL;
        CFIndex namelen = strlen(file.cFileName);
        if (file.cFileName[0] == '.' && (namelen == 1 || (namelen == 2  && file.cFileName[1] == '.'))) {
            continue;
        }
        if (extLen > 0) {
            // Check to see if it matches the extension we're looking for.
            if (_stricmp(&(file.cFileName[namelen - extLen]), extBuff) != 0) {
                continue;
            }
        }
        if (base == NULL) {
            base = CFURLCreateFromFileSystemRepresentation(alloc, path, pathLength, true);
            releaseBase = true;
        }
        // MF:!!! What about the trailing slash?
        fileURL = CFURLCreateFromFileSystemRepresentationRelativeToBase(alloc, file.cFileName, namelen, false, base);
        CFArrayAppendValue(files, fileURL);
        CFRelease(fileURL);
    } while (FindNextFileA(handle, &file));
    FindClose(handle);
    path[pathLength] = '\0';

#elif defined(__svr4__) || defined(__hpux__) || defined(__LINUX__) || defined(__FREEBSD__)
    /* Solaris and HPUX Implementation */

    dirp = opendir(dirPath);
    if (!dirp) {
        if (extension) {
            CFRelease(extension);
        }
        return NULL;
        // raiseErrno("opendir", path);
    }
    files = CFArrayCreateMutable(alloc, 0, & kCFTypeArrayCallBacks);

    while((dp = readdir(dirp)) != NULL) {
        CFURLRef fileURL;
	unsigned namelen = strlen(dp->d_name);

        // skip . & ..; they cause descenders to go berserk
	if (dp->d_name[0] == '.' && (namelen == 1 || (namelen == 2 && dp->d_name[1] == '.'))) {
            continue;
        }

        if (extLen > 0) {
            // Check to see if it matches the extension we're looking for.
            if (strncmp(&(dp->d_name[namelen - extLen]), extBuff, extLen) != 0) {
                continue;
            }
        }
        if (dirURL == NULL) {
            dirURL = CFURLCreateFromFileSystemRepresentation(alloc, dirPath, pathLength, true);
            releaseBase = true;
        }
        // MF:!!! What about the trailing slash?
        fileURL = CFURLCreateFromFileSystemRepresentationRelativeToBase(alloc, dp->d_name, namelen, false, dirURL);
        CFArrayAppendValue(files, fileURL);
        CFRelease(fileURL);
    }
    err = closedir(dirp);
    if (err != 0) {
        CFRelease(files);
        if (releaseBase) {
            CFRelease(dirURL);
        }
        if (extension) {
            CFRelease(extension);
        }
        return NULL;
        // raiseErrno("closedir", path);
    }
    
#elif defined(__MACH__)
    /* Mac OS X Variables - repeated for convenience */
    // int fd, numread;
    // long basep;
    // char dirge[8192];
    // UInt8 pathBuf[CFMaxPathSize];
    /* Mac OS X Implementation */

    if (!dirPath) {
        if (!CFURLGetFileSystemRepresentation(dirURL, true, pathBuf, CFMaxPathLength)) {
            if (extension) CFRelease(extension);
            return NULL;
        } else {
            dirPath = pathBuf;
            pathLength = strlen(dirPath);
        }
    }
    fd = open(dirPath, O_RDONLY, 0777);
    if (fd < 0) {
        if (extension) {
            CFRelease(extension);
        }
        return NULL;
    }
    files = CFArrayCreateMutable(alloc, 0, &kCFTypeArrayCallBacks);

    while ((numread = getdirentries(fd, dirge, sizeof(dirge), &basep)) > 0) {
        struct dirent *dent;
        for (dent = (struct dirent *)dirge; dent < (struct dirent *)(dirge + numread); dent = (struct dirent *)((char *)dent + dent->d_reclen)) {
            CFURLRef fileURL;
            CFIndex nameLen;

            nameLen = dent->d_namlen;
            // skip . & ..; they cause descenders to go berserk
            if (0 == dent->d_fileno || (dent->d_name[0] == '.' && (nameLen == 1 || (nameLen == 2 && dent->d_name[1] == '.')))) {
                continue;
            }
            if (extLen > 0) {
                // Check to see if it matches the extension we're looking for.
                if (strncmp(&(dent->d_name[nameLen - extLen]), extBuff, extLen) != 0) {
                    continue;
                }
            }
            if (dirURL == NULL) {
                dirURL = CFURLCreateFromFileSystemRepresentation(alloc, dirPath, pathLength, true);
                releaseBase = true;
            }

            if (dent->d_type == DT_DIR || dent->d_type == DT_UNKNOWN) {
                Boolean isDir = (dent->d_type == DT_DIR);
                if (!isDir) {
                    // Ugh; must stat.
                    char subdirPath[CFMaxPathLength];
                    struct stat statBuf;
                    strncpy(subdirPath, dirPath, pathLength);
                    subdirPath[pathLength] = '/';
                    strncpy(subdirPath + pathLength + 1, dent->d_name, nameLen);
                    subdirPath[pathLength + nameLen + 1] = '\0';
                    if (stat(subdirPath, &statBuf) == 0) {
                        isDir = ((statBuf.st_mode & S_IFMT) == S_IFDIR);
                    }
                }
                fileURL = CFURLCreateFromFileSystemRepresentationRelativeToBase(alloc, dent->d_name, nameLen, isDir, dirURL);
            } else {
                fileURL = CFURLCreateFromFileSystemRepresentationRelativeToBase (alloc, dent->d_name, nameLen, false, dirURL);
            }
            CFArrayAppendValue(files, fileURL);
            CFRelease(fileURL);
        }
    }
    close(fd);
    if  (-1 == numread) {
        CFRelease(files);
        if (releaseBase) {
            CFRelease(dirURL);
        }
        if (extension) {
            CFRelease(extension);
        }
        return NULL;
    }
#elif defined(__MACOS8__)
    /* Mac OS 8 Implementation */
    if (!dirURL) {
        if (dirPath) {
            dirURL = CFURLCreateFromFileSystemRepresentation(alloc, dirPath, pathLength, true);
        } else {
            dirURL = _CFCreateURLFromFSSpec(alloc, dirSpec, true);
        }
        releaseBase = true;
    }
    if (!dirURL || (!dirSpec && !_CFGetFSSpecFromURL(alloc, dirURL, &specBuf))) {
        if (extension) CFRelease(extension);
        if (dirURL && releaseBase) CFRelease(dirURL);
        return NULL;
    } else {
        CInfoPBRec catInfoRec;
        OSErr oserr;
        int16_t vRefNum;
        FSSpec *dirSp;

        if (!dirSpec) {
            dirSp = &specBuf;
        } else {
        	dirSp = (FSSpec *)dirSpec;
        }
        vRefNum = dirSp->vRefNum;

        // First get the dir id of the directory
        catInfoRec.dirInfo.ioCompletion = NULL;
        catInfoRec.dirInfo.ioNamePtr = dirSp->name;
        catInfoRec.dirInfo.ioVRefNum = vRefNum;
        catInfoRec.dirInfo.ioFDirIndex = 0;
        catInfoRec.dirInfo.ioDrDirID = dirSp->parID;
        oserr = PBGetCatInfo(&catInfoRec, false);
        
        if (oserr == noErr) {
            Boolean done = false;
            SInt32 dirIndex = 1;
            SInt32 dirID = catInfoRec.dirInfo.ioDrDirID;
            Str255 namePtr;
            
            files = CFArrayCreateMutable(alloc, 0, &kCFTypeArrayCallBacks);

            catInfoRec.dirInfo.ioCompletion = NULL;
            catInfoRec.dirInfo.ioVRefNum = vRefNum;

            while (!done) {
                catInfoRec.dirInfo.ioNamePtr = namePtr;
                catInfoRec.dirInfo.ioFDirIndex = dirIndex;
                catInfoRec.dirInfo.ioDrDirID = dirID;
                oserr = PBGetCatInfo(&catInfoRec, false);
                if (oserr != noErr) {
                    done = true;
                } else {
                    CFMutableStringRef str = CFStringCreateMutable(alloc, 0);
                    dirIndex++;
                    if (str) {
                    	CFStringAppendCString(str, ":", kCFStringEncodingASCII);
                    	CFStringAppendPascalString(str, namePtr, CFStringFileSystemEncoding());
                   		if (!extension || CFStringHasSuffix(str, extension)) {
	                        CFURLRef url = CFURLCreateWithFileSystemPathRelativeToBase(alloc, str, kCFURLHFSPathStyle, (catInfoRec.dirInfo.ioFlAttrib & (1<<4)) ? true : false, dirURL);
                        	if (url) {
                            	CFArrayAppendValue(files, url);
                            	CFRelease(url);
                        	}
                        }
                        CFRelease(str);
                    }
                }
            }
        }
    }
#else
    
#error _CFContentsOfDirectory() unknown architechture, not implemented
    
#endif

    if (extension) {
        CFRelease(extension);
    }
    if (releaseBase) {
        CFRelease(dirURL);
    }
    return files;
}

__private_extern__ SInt32 _CFGetFileProperties(CFAllocatorRef alloc, CFURLRef pathURL, Boolean *exists, SInt32 *posixMode, int64_t *size, CFDateRef *modTime, SInt32 *ownerID, CFArrayRef *dirContents) {
    Boolean fileExists;
    Boolean isDirectory = false;
#if !defined(__MACOS8__)
    struct stat statBuf;
    char path[CFMaxPathLength];
#else
    FSSpec spec;
    OSStatus errCode;
    CInfoPBRec fileParams;
#endif
    if ((exists == NULL) && (posixMode == NULL) && (size == NULL) && (modTime == NULL) && (ownerID == NULL) && (dirContents == NULL)) {
        // Nothing to do.
        return 0;
    }
#if !defined(__MACOS8__)
    if (!CFURLGetFileSystemRepresentation(pathURL, true, path, CFMaxPathLength)) {
        return -1;
    }

    if (stat(path, &statBuf) != 0) {
        // stat failed, but why?
        if (thread_errno() == ENOENT) {
            fileExists = false;
        } else {
            return thread_errno();
        }
    } else {
        fileExists = true;
        isDirectory = ((statBuf.st_mode & S_IFMT) == S_IFDIR);
    }
#else
    if (!_CFGetFSSpecFromURL(alloc, pathURL, &spec)) {
        return -1;
    }
    fileParams.hFileInfo.ioNamePtr = &spec.name;
    fileParams.hFileInfo.ioVRefNum = spec.vRefNum;
    fileParams.hFileInfo.ioFDirIndex = 0;
    fileParams.hFileInfo.ioDirID = spec.parID;
    errCode = PBGetCatInfo(&fileParams, false);
    if (errCode != noErr) {
        if (errCode == fnfErr) {
            fileExists = false;
        } else {
            return -1;
        }
    } else {
        fileExists = true;
        isDirectory = ((fileParams.hFileInfo.ioFlAttrib & ioDirMask) != 0);
    }
#endif

    if (exists != NULL) {
        *exists = fileExists;
    }

    if (posixMode != NULL) {
        if (fileExists) {
#if !defined(__MACOS8__)
            *posixMode = statBuf.st_mode;
#else
            // These are the values for S_IFDIR and S_IFREG from stat.h.  We can't simply include stat.h because it's not available on 8/9, so we grab the UNIX values from the header on X.  REW, 6/5/2000
            *posixMode = (isDirectory ? 0040000 : 0100000) | 0777;
#endif
        } else {
            *posixMode = 0;
        }
    }

    if (size != NULL) {
        if (fileExists) {
#if !defined(__MACOS8__)
            *size = statBuf.st_size;
#else
            *size = isDirectory ? 0 : fileParams.hFileInfo.ioFlLgLen;
#endif
        } else {
            *size = 0;
        }
    }

    if (modTime != NULL) {
        if (fileExists) {
            CFTimeInterval theTime;
#if !defined(__MACOS8__)
            theTime = kCFAbsoluteTimeIntervalSince1970 + statBuf.st_mtime;
#else
            if (isDirectory) {
                theTime = kCFAbsoluteTimeIntervalSince1904 + fileParams.dirInfo.ioDrMdDat;
            } else {
                theTime = kCFAbsoluteTimeIntervalSince1904 + fileParams.hFileInfo.ioFlMdDat;
            }
#warning Should correct for time zone here REW, 6/5/2000
#endif
            *modTime = CFDateCreate(alloc, theTime);
        } else {
            *modTime = NULL;
        }
    }

    if (ownerID != NULL) {
        if (fileExists) {
#if !defined(__MACOS8__)
            *ownerID = statBuf.st_uid;
#else
            *ownerID = -1; // ownerID not supported on MacOS 8/9
#endif
        } else {
            *ownerID = -1;
        }
    }
    
    if (dirContents != NULL) {
        if (fileExists && isDirectory) {
#if !defined(__MACOS8__)
            CFMutableArrayRef contents = _CFContentsOfDirectory(alloc, path, NULL, pathURL, NULL);
#else
#warning Check me
            CFMutableArrayRef contents = _CFContentsOfDirectory(alloc, NULL, &spec, pathURL, NULL);
#endif
            if (contents) {
                *dirContents = contents;
            } else {
                *dirContents = NULL;
            }
        } else {
            *dirContents = NULL;
        }
    }
    return 0;
}


// MF:!!! Should pull in the rest of the UniChar based path utils from Foundation.
#if defined(__MACH__) || defined(__svr4__) || defined(__hpux__) || defined(__LINUX__) || defined(__FREEBSD__)
    #define UNIX_PATH_SEMANTICS
#elif defined(__WIN32__)
    #define WINDOWS_PATH_SEMANTICS
#elif defined(__MACOS8__)
    #define HFS_PATH_SEMANTICS
#else
#error Unknown platform
#endif

#if defined(WINDOWS_PATH_SEMANTICS)
    #define CFPreferredSlash	((UniChar)'\\')
#elif defined(UNIX_PATH_SEMANTICS)
    #define CFPreferredSlash	((UniChar)'/')
#elif defined(HFS_PATH_SEMANTICS)
    #define CFPreferredSlash	((UniChar)':')
#else
    #error Cannot define NSPreferredSlash on this platform
#endif

#if defined(HFS_PATH_SEMANTICS)
#define HAS_DRIVE(S) (false)
#define HAS_NET(S) (false)
#else
#define HAS_DRIVE(S) ((S)[1] == ':' && (('A' <= (S)[0] && (S)[0] <= 'Z') || ('a' <= (S)[0] && (S)[0] <= 'z')))
#define HAS_NET(S) ((S)[0] == '\\' && (S)[1] == '\\')
#endif

#if defined(WINDOWS_PATH_SEMANTICS)
    #define IS_SLASH(C)	((C) == '\\' || (C) == '/')
#elif defined(UNIX_PATH_SEMANTICS)
    #define IS_SLASH(C)	((C) == '/')
#elif defined(HFS_PATH_SEMANTICS)
    #define IS_SLASH(C)	((C) == ':')
#endif

__private_extern__ Boolean _CFIsAbsolutePath(UniChar *unichars, CFIndex length) {
    if (length < 1) {
        return false;
    }
#if defined(WINDOWS_PATH_SEMANTICS)
    if (unichars[0] == '~') {
        return true;
    }
    if (length < 2) {
        return false;
    }
    if (HAS_NET(unichars)) {
        return true;
    }
    if (length < 3) {
        return false;
    }
    if (IS_SLASH(unichars[2]) && HAS_DRIVE(unichars)) {
        return true;
    }
#elif defined(HFS_PATH_SEMANTICS)
    return !IS_SLASH(unichars[0]);
#else
    if (unichars[0] == '~') {
        return true;
    }
    if (IS_SLASH(unichars[0])) {
        return true;
    }
#endif
    return false;
}

__private_extern__ Boolean _CFStripTrailingPathSlashes(UniChar *unichars, CFIndex *length) {
    Boolean destHasDrive = (1 < *length) && HAS_DRIVE(unichars);
    CFIndex oldLength = *length;
    while (((destHasDrive && 3 < *length) || (!destHasDrive && 1 < *length)) && IS_SLASH(unichars[*length - 1])) {
        (*length)--;
    }
    return (oldLength != *length);
}

__private_extern__ Boolean _CFAppendPathComponent(UniChar *unichars, CFIndex *length, CFIndex maxLength, UniChar *component, CFIndex componentLength) {
    if (0 == componentLength) {
        return true;
    }
    if (maxLength < *length + 1 + componentLength) {
        return false;
    }
    switch (*length) {
    case 0:
        break;
    case 1:
        if (!IS_SLASH(unichars[0])) {
            unichars[(*length)++] = CFPreferredSlash;
        }
        break;
    case 2:
        if (!HAS_DRIVE(unichars) && !HAS_NET(unichars)) {
            unichars[(*length)++] = CFPreferredSlash;
        }
        break;
    default:
        unichars[(*length)++] = CFPreferredSlash;
        break;
    }
    memmove(unichars + *length, component, componentLength * sizeof(UniChar));
    *length += componentLength;
    return true;
}

__private_extern__ Boolean _CFAppendPathExtension(UniChar *unichars, CFIndex *length, CFIndex maxLength, UniChar *extension, CFIndex extensionLength) {
    if (maxLength < *length + 1 + extensionLength) {
        return false;
    }
    if ((0 < extensionLength && IS_SLASH(extension[0])) || (1 < extensionLength && HAS_DRIVE(extension))) {
        return false;
    }
    _CFStripTrailingPathSlashes(unichars, length);
    switch (*length) {
    case 0:
        return false;
    case 1:
        if (IS_SLASH(unichars[0]) || unichars[0] == '~') {
            return false;
        }
        break;
    case 2:
        if (HAS_DRIVE(unichars) || HAS_NET(unichars)) {
            return false;
        }
        break;
    case 3:
        if (IS_SLASH(unichars[2]) && HAS_DRIVE(unichars)) {
            return false;
        }
        break;
    }
    if (0 < *length && unichars[0] == '~') {
        CFIndex idx;
        Boolean hasSlash = false;
        for (idx = 1; idx < *length; idx++) {
            if (IS_SLASH(unichars[idx])) {
                hasSlash = true;
                break;
            }
        }
        if (!hasSlash) {
            return false;
        }
    }
    unichars[(*length)++] = '.';
    memmove(unichars + *length, extension, extensionLength * sizeof(UniChar));
    *length += extensionLength;
    return true;
}

__private_extern__ Boolean _CFTransmutePathSlashes(UniChar *unichars, CFIndex *length, UniChar replSlash) {
    CFIndex didx, sidx, scnt = *length;
    sidx = (1 < *length && HAS_NET(unichars)) ? 2 : 0;
    didx = sidx;
    while (sidx < scnt) {
        if (IS_SLASH(unichars[sidx])) {
            unichars[didx++] = replSlash;
            for (sidx++; sidx < scnt && IS_SLASH(unichars[sidx]); sidx++);
        } else {
            unichars[didx++] = unichars[sidx++];
        }
    }
    *length = didx;
    return (scnt != didx);
}

__private_extern__ CFIndex _CFStartOfLastPathComponent(UniChar *unichars, CFIndex length) {
    CFIndex idx;
    if (length < 2) {
        return 0;
    }
    for (idx = length - 1; idx; idx--) {
        if (IS_SLASH(unichars[idx - 1])) {
            return idx;
        }
    }
    if ((2 < length) && HAS_DRIVE(unichars)) {
        return 2;
    }
    return 0;
}

__private_extern__ CFIndex _CFLengthAfterDeletingLastPathComponent(UniChar *unichars, CFIndex length) {
    CFIndex idx;
    if (length < 2) {
        return 0;
    }
    for (idx = length - 1; idx; idx--) {
        if (IS_SLASH(unichars[idx - 1])) {
            if ((idx != 1) && (!HAS_DRIVE(unichars) || idx != 3)) {
                return idx - 1;
            }
            return idx;
        }
    }
    if ((2 < length) && HAS_DRIVE(unichars)) {
        return 2;
    }
    return 0;
}

__private_extern__ CFIndex _CFStartOfPathExtension(UniChar *unichars, CFIndex length) {
    CFIndex idx;
    if (length < 2) {
        return 0;
    }
    for (idx = length - 1; idx; idx--) {
        if (IS_SLASH(unichars[idx - 1])) {
            return 0;
        }
        if (unichars[idx] != '.') {
            continue;
        }
        if (idx == 2 && HAS_DRIVE(unichars)) {
            return 0;
        }
        return idx;
    }
    return 0;
}

__private_extern__ CFIndex _CFLengthAfterDeletingPathExtension(UniChar *unichars, CFIndex length) {
    unsigned start = _CFStartOfPathExtension(unichars, length);
    return ((0 < start) ? start : length);
}

#undef ENOENT
#undef CF_OPENFLGS
#undef UNIX_PATH_SEMANTICS
#undef WINDOWS_PATH_SEMANTICS
#undef HFS_PATH_SEMANTICS
#undef CFPreferredSlash
#undef HAS_DRIVE
#undef HAS_NET
#undef IS_SLASH

