/*
 * Copyright (c) 2001 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*	CFPlatform.c
	Copyright 1999-2001, Apple, Inc. All rights reserved.
	Responsibility: Christopher Kane
*/

#include "CFPlatform.h"
#include "CFFileUtilities.h"
#include "CFInternal.h"
#include "CFPriv.h"
#if defined(__MACOS8__)
    #include <Timer.h>
    #include <stat.h>
    #include <string.h>
    #include <Processes.h>
    #include <unistd.h>
#elif defined(__WIN32__)
    #include <sys/stat.h>
    #include <string.h>
    #include <windows.h>
    #include <stdlib.h>
#else
    #include <sys/stat.h>
    #include <string.h>
    #include <unistd.h>
    #include <pwd.h>
#endif

extern char *getenv(const char *name);

#if defined(__MACH__)
#define kCFPlatformInterfaceStringEncoding	kCFStringEncodingUTF8
#else
#define kCFPlatformInterfaceStringEncoding	CFStringGetSystemEncoding()
#endif __MACH__

char **_CFArgv(void) {
#if defined(__MACOS8__)
    // On Mac OS 8 just manufacture a dummy argv
    static char *bogusArgs[] = { "UnknownApp", NULL };
    return bogusArgs;
#elif defined(__svr4__) || defined(__hpux__)
    if (NULL == NXArgv) {
        static char *bogusArgs[] = { "UnknownApp", NULL };
        return bogusArgs;
    }
    return NXArgv;
#elif defined(__MACH__)
    extern char ***_NSGetArgv(void);
    return *_NSGetArgv();
#elif defined(__WIN32__)
    return __argv;
#elif defined(__FREEBSD__) || defined(__LINUX__)
#warning (CF) FreeBSD and Linux bogus for _CFArgv
    static char *bogusArgs[] = { "UnknownApp", NULL };
    return bogusArgs;
#else
#warning (CF) Unknown platform for _CFArgv, _CFArgc
    static char *bogusArgs[] = { "UnknownApp", NULL };
    return bogusArgs;
#endif
}

int _CFArgc(void) {
#if defined(__MACOS8__)
    // On Mac OS 8 the dummy argc is 1
    return 1;
#elif defined(__svr4__) || defined(__hpux__)
    if (NULL == NXArgv) {
        return 1;
    }
    return NXArgc;
#elif defined(__MACH__)
    extern int *_NSGetArgc(void);
    return *_NSGetArgc();
#elif defined(__WIN32__)
    return __argc;
#else
    return 1;
#endif
}

#if defined(__MACOS8__)
__private_extern__ char *getenv(void) {
    return NULL;
}
#endif

#if !defined(__MACOS8__)
__private_extern__ Boolean _CFGetCurrentDirectory(char *path, int maxlen) {
#if defined(__WIN32__)
    DWORD len = GetCurrentDirectoryA(maxlen, path);
    return (0 != len && len + 1 <= maxlen);
#else
    return getcwd(path, maxlen) != NULL;
#endif
}
#endif

#if !defined(__MACOS8__)

#if defined(__WIN32__)
#define PATH_LIST_SEP ';'
#else
#define PATH_LIST_SEP ':'
#endif

static char *_CFSearchForNameInPath(CFAllocatorRef alloc, const char *name, char *path) {
    struct stat statbuf;
    char *nname = CFAllocatorAllocate(alloc, strlen(name) + strlen(path) + 2, 0);
    for (;;) {
        char *p = (char *)strchr(path, PATH_LIST_SEP);
        if (NULL != p) {
            *p = '\0';
        }
        nname[0] = '\0';
        strcat(nname, path);
        strcat(nname, "/");
        strcat(nname, name);
        // Could also do access(us, X_OK) == 0 in next condition,
        // for executable-only searching
        if (0 == stat(nname, &statbuf) && (statbuf.st_mode & S_IFMT) == S_IFREG) {
            if (p != NULL) {
                *p = PATH_LIST_SEP;
            }
            return nname;
        }
        if (NULL == p) {
            break;
        }
        *p = PATH_LIST_SEP;
        path = p + 1;
    }
    CFAllocatorDeallocate(alloc, nname);
    return NULL;
}

#endif

static const char *__CFProcessPath = NULL;
static const char *__CFProcessName = NULL;
static CFStringRef __CFProcessNameString = NULL;

__private_extern__ const char *_CFProcessPath(void) {
    CFAllocatorRef alloc = NULL;
    char *thePath = NULL;
    
    if (!__CFProcessPath) {
        thePath = getenv("CFProcessPath");
        
        alloc = CFRetain(__CFGetDefaultAllocator());
        
	if (thePath) {
	    int len = strlen(thePath);
	    __CFProcessPath = CFAllocatorAllocate(alloc, len+1, 0);
	    memmove((char *)__CFProcessPath, thePath, len + 1);
	}
    }
    
    if (!__CFProcessPath) {
	char buf[CFMaxPathSize] = {0};
#if defined(__WIN32__)
	HINSTANCE hinst = GetModuleHandle(NULL);
	DWORD rlen = hinst ? GetModuleFileName(hinst, buf, 1028) : 0;
	thePath = rlen ? buf : NULL;
#elif defined(__MACOS8__)
        // This gets both the path and the name.
        ProcessSerialNumber gProcessID;
        ProcessInfoRec processInfo;
        Str255 processName;
        FSSpec processAppSpec;
        
        processInfo.processInfoLength = sizeof(ProcessInfoRec);
        processInfo.processName = processName;
        processInfo.processAppSpec = &processAppSpec;

        if ((GetCurrentProcess(&gProcessID) == noErr) && (GetProcessInformation(&gProcessID, &processInfo) == noErr)) {
            int len;
            CFURLRef url;
            CFStringRef pathStr;
            
            // Get the path
            url = _CFCreateURLFromFSSpec(alloc, (void *)(&processAppSpec), false);
            pathStr = url ? CFURLCopyFileSystemPath(url, kCFURLPOSIXPathStyle): NULL;
            if (url) CFRelease(url);
            if ((pathStr != NULL) && (CFStringGetLength(pathStr) > 0)) {
            	__CFProcessPath = CFAllocatorAllocate(alloc, CFMaxPathSize, 0);
            	if (!CFStringGetCString(pathStr, (char *)__CFProcessPath, CFMaxPathSize, CFStringFileSystemEncoding())) {
            	    CFAllocatorDeallocate(alloc, (char *)__CFProcessPath);
            	    __CFProcessPath = "";
            	}
            	CFRelease(pathStr);
            } else {
            	__CFProcessPath = "";
            	if (pathStr) CFRelease(pathStr);
            }

            // Get the name
            len = StrLength(processName);
            if (len > 0) {
            	__CFProcessName = CFAllocatorAllocate(alloc, len + 1, 0);
            	memmove((char *)__CFProcessName, processName + 1, len);
            	((char *)__CFProcessName)[len] = '\0';	// terminate string
            } else {
            	__CFProcessName = "";
            }
        } else {
            __CFProcessPath = "";
            __CFProcessName = "";
        }
        thePath = (char *)__CFProcessPath;
#else
	struct stat statbuf;
        const char *arg0 = _CFArgv()[0];
        if (arg0[0] == '/') {
            // We've got an absolute path; look no further;
            thePath = (char *)arg0;
        } else {
            char *theList = getenv("PATH");
            if (NULL != theList && NULL == strrchr(arg0, '/')) {
                thePath = _CFSearchForNameInPath(alloc, arg0, theList);
                if (thePath) {
                    // User could have "." or "../bin" or other relative path in $PATH
                    if (('/' != thePath[0]) && _CFGetCurrentDirectory(buf, CFMaxPathSize)) {
                        strcat(buf, "/");
                        strcat(buf, thePath);
                        if (0 == stat(buf, &statbuf)) {
                            CFAllocatorDeallocate(alloc, (void *)thePath);
                            thePath = buf;
                        }
                    }
                    if (thePath != buf) {
                        strcpy(buf, thePath);
                        CFAllocatorDeallocate(alloc, (void *)thePath);
                        thePath = buf;
                    }
                }
            }
        }
        
	// After attempting a search through $PATH, if existant,
	// try prepending the current directory to argv[0].
        if (!thePath && _CFGetCurrentDirectory(buf, CFMaxPathSize)) {
            if (buf[strlen(buf)-1] != '/') {
                strcat(buf, "/");
            }
	    strcat(buf, arg0);
            if (0 == stat(buf, &statbuf)) {
		thePath = buf;
            }
	}

        if (thePath) {
            // We are going to process the buffer replacing all "/./" with "/"
            CFIndex srcIndex = 0, dstIndex = 0;
            CFIndex len = strlen(thePath);
            for (srcIndex=0; srcIndex<len; srcIndex++) {
                thePath[dstIndex] = thePath[srcIndex];
                dstIndex++;
                if ((srcIndex < len-2) && (thePath[srcIndex] == '/') && (thePath[srcIndex+1] == '.') && (thePath[srcIndex+2] == '/')) {
                    // We are at the first slash of a "/./"  Skip the "./"
                    srcIndex+=2;
                }
            }
            thePath[dstIndex] = 0;
        }
#endif
        if (!thePath) {
	    thePath = _CFArgv()[0];
        }
	if (thePath) {
	    int len = strlen(thePath);
	    __CFProcessPath = CFAllocatorAllocate(alloc, len + 1, 0);
            memmove((char *)__CFProcessPath, thePath, len + 1);
	}
    }
    if (!__CFProcessPath) {
	__CFProcessPath = "";
    }
    return __CFProcessPath;
}

__private_extern__ const char *_CFProcessName(void) {
    if (!__CFProcessName) {
        CFAllocatorRef alloc = CFRetain(__CFGetDefaultAllocator());

        char *theName = NULL, *fullpath = (char *)_CFProcessPath();
        int len;
        
#if defined(__WIN32__)
        WIN32_FIND_DATA findData;
        HANDLE findHandle = FindFirstFileA(fullpath, &findData);
        if (INVALID_HANDLE_VALUE == findHandle) {
            char *ch;
            for (ch = fullpath; *ch; ch++) {
                if ((*ch == '/') || (*ch == '\\') || (*ch == ':')) {
                    theName = ch + 1;
                }
            }
        } else {
            theName = findData.cFileName;
        }
        if (!theName || 0 == strlen(theName)) {
	    theName = fullpath;
        }
	len = strlen(theName);
	__CFProcessName = CFAllocatorAllocate(alloc, len + 1, 0);
        memmove((char *)__CFProcessName, theName, len + 1);
        if (_stricmp(__CFProcessName + len - 4, ".exe") == 0) {
	    *((char *)__CFProcessName + len - 4) = '\0';
        }
        if (_stricmp(__CFProcessName + len - 5, ".exe.") == 0) {
	    *((char *)__CFProcessName + len - 5) = '\0';
        }
        if (INVALID_HANDLE_VALUE != findHandle) {
            FindClose(findHandle);
        }
#elif defined(__MACOS8__)
        // Nothing to do here since the previous call to _CFProcessPath() will have already done the work.
#else
        do {
            if (*fullpath == '/') {
                theName = (char *)fullpath;
            }
        } while (*++fullpath);
	theName = (theName ? theName + 1 : (char *)_CFProcessPath());
	len = strlen(theName);
        __CFProcessName = CFAllocatorAllocate(alloc, len + 1, 0);
        memmove((char *)__CFProcessName, theName, len + 1);
#endif
    }
    if (!__CFProcessName) {
	__CFProcessName = "";
    }
    return __CFProcessName;
}

__private_extern__ CFStringRef _CFProcessNameString(void) {
    if (!__CFProcessNameString) {
        const char *processName = _CFProcessName();
        if (!processName) processName = "";
        __CFProcessNameString = CFStringCreateWithCString(__CFGetDefaultAllocator(), _CFProcessName(), kCFPlatformInterfaceStringEncoding);
    }
    return __CFProcessNameString;
}

static CFURLRef __CFHomeDirectory = NULL;
static CFStringRef __CFUserName = NULL;
static int __CFEUID = -1;
static int __CFUID = -1;

#if defined(__MACH__) || defined(__svr4__) || defined(__hpux__) || defined(__LINUX__) || defined(__FREEBSD__)
static CFURLRef _CFCopyHomeDirURLForUser(struct passwd *upwd) {
    CFURLRef home = NULL;
    if (upwd && upwd->pw_dir) {
        home = CFURLCreateFromFileSystemRepresentation(NULL, upwd->pw_dir, strlen(upwd->pw_dir), true);
    }
    return home;
}

static void _CFUpdateUserInfo(void) {
    struct passwd *upwd;

    __CFEUID = geteuid();
    __CFUID = getuid();
    if (__CFHomeDirectory)  CFRelease(__CFHomeDirectory);
    __CFHomeDirectory = NULL;
    if (__CFUserName) CFRelease(__CFUserName);
    __CFUserName = NULL;

    upwd = getpwuid(__CFEUID ? __CFEUID : __CFUID);
    __CFHomeDirectory = _CFCopyHomeDirURLForUser(upwd);
    if (!__CFHomeDirectory) {
        const char *cpath = getenv("HOME");
        if (cpath) {
            __CFHomeDirectory = CFURLCreateFromFileSystemRepresentation(NULL, cpath, strlen(cpath), true);
        }
    }

    // This implies that UserManager stores directory info in CString
    // rather than FileSystemRep.  Perhaps this is wrong & we should
    // expect NeXTSTEP encodings.  A great test of our localized system would
    // be to have a user "O-umlat z e r".  XXX
    if (upwd && upwd->pw_name) {
        __CFUserName = CFStringCreateWithCString(NULL, upwd->pw_name, kCFPlatformInterfaceStringEncoding);
    } else {
        const char *cuser = getenv("USER");
        if (cuser)
            __CFUserName = CFStringCreateWithCString(NULL, cuser, kCFPlatformInterfaceStringEncoding);
    }
}
#endif

CFURLRef _CFCreateHomeDirectoryURLForUser(CFStringRef uName) {
#if defined(__MACH__) || defined(__svr4__) || defined(__hpux__) || defined(__LINUX__) || defined(__FREEBSD__)
    if (!uName) {
        if (geteuid() != __CFEUID || getuid() != __CFUID || !__CFHomeDirectory)
            _CFUpdateUserInfo();
        if (__CFHomeDirectory) CFRetain(__CFHomeDirectory);
        return __CFHomeDirectory;
    } else {
        struct passwd *upwd = NULL;
        char buf[128], *user;
        SInt32 len = CFStringGetLength(uName), size = CFStringGetMaximumSizeForEncoding(len, kCFPlatformInterfaceStringEncoding);
        CFIndex usedSize;
        if (size < 127) {
            user = buf;
        } else {
            user = CFAllocatorAllocate(kCFAllocatorDefault, size+1, 0);
        }
        if (CFStringGetBytes(uName, CFRangeMake(0, len), kCFPlatformInterfaceStringEncoding, 0, true, user, size, &usedSize) == len) {
            user[usedSize] = '\0';
            upwd = getpwnam(user);
        }
        if (buf != user) {
            CFAllocatorDeallocate(kCFAllocatorDefault, user);
        }
        return _CFCopyHomeDirURLForUser(upwd);
    }
#elif defined(__WIN32__)
#warning CF: Windows home directory goop disabled
    return NULL;
#if 0
    CFString *user = !uName ? CFUserName() : uName;

    if (!uName || CFEqual(user, CFUserName())) {
        const char *cpath = getenv("HOMEPATH");
        const char *cdrive = getenv("HOMEDRIVE");
        if (cdrive && cpath) {
            char fullPath[CFMaxPathSize];
            CFStringRef str;
            strcpy(fullPath, cdrive);
            strncat(fullPath, cpath, CFMaxPathSize-strlen(cdrive)-1);
            str = CFStringCreateWithCString(NULL, fullPath, kCFPlatformInterfaceStringEncoding);
            home = CFURLCreateWithFileSystemPath(NULL, str, kCFURLWindowsPathStyle, true);
            CFRelease(str);
        }
    }
    if (!home) {
        struct _USER_INFO_2 *userInfo;
        HINSTANCE hinstDll = GetModuleHandleA("NETAPI32");
        if (!hinstDll)
            hinstDll = LoadLibraryEx("NETAPI32", NULL, 0);
        if (hinstDll) {
            FARPROC lpfn = GetProcAddress(hinstDll, "NetUserGetInfo");
            if (lpfn) {
                unsigned namelen = CFStringGetLength(user);
                UniChar *username;
                username = CFAllocatorAllocate(kCFAllocatorDefault, sizeof(UniChar) * (namelen + 1), 0);
                CFStringGetCharacters(user, CFRangeMake(0, namelen), username);
                if (!(*lpfn)(NULL, (LPWSTR)username, 2, (LPBYTE *)&userInfo)) {
                    UInt32 len = 0;
                    CFMutableStringRef str;
                    while (userInfo->usri2_home_dir[len] != 0) len ++;
                    str = CFStringCreateMutable(NULL, len+1);
                    CFStringAppendCharacters(str, userInfo->usri2_home_dir, len);
                    home = CFURLCreateWithFileSystemPath(NULL, str, kCFURLWindowsPathStyle, true);
                    CFRelease(str);
                }
                CFAllocatorDeallocate(kCFAllocatorDefault, username);
            }
        } else {
        }
    }
    // We could do more here (as in KB Article Q101507). If that article is to
    // be believed, we should only run into this case on Win95, or through
    // user error.
    if (CFStringGetLength(CFURLGetPath(home)) == 0) {
        CFRelease(home);
        home=NULL;
    }
#endif

#elif defined(__MACOS8__)
    // No home directories on MacOS 8.x
    return NULL;
#else
#error Dont know how to compute users home directories on this platform
#endif
}

__private_extern__ CFURLRef _CFHomeDirectory(void) {
#if defined(__MACH__) || defined(__svr4__) || defined(__hpux__) || defined(__LINUX__) || defined(__FREEBSD__)
    CFURLRef url = _CFCreateHomeDirectoryURLForUser(NULL);
    if (url) CFRelease(url); // because this is the cached home directory for the current user, it's already retained once in the cache; we don't want it retained again (as _CFCreateHomeDirectoryURLForUser() has just done) -- REW, 3/1/2000
    return url;
#elif defined(__MACOS8__)
    // No home directories under MacOS 8.x
    return NULL;
#endif
}

__private_extern__ CFStringRef _CFUserName(void) {
#if defined(__MACH__) || defined(__svr4__) || defined(__hpux__) || defined(__LINUX__) || defined(__FREEBSD__)
    if (geteuid() != __CFEUID || getuid() != __CFUID)
	_CFUpdateUserInfo();
#elif defined(__WIN32__)
    if (!__CFUserName) {
	char username[1040];
	DWORD size = 1040;
	username[0] = 0;
	if (GetUserNameA(username, &size)) {
            __CFUserName = CFStringCreateWithCString(NULL, username, kCFPlatformInterfaceStringEncoding);
	} else {
	    const char *cname = getenv("USERNAME");
	    if (cname)
                __CFUserName = CFStringCreateWithCString(NULL, cname, kCFPlatformInterfaceStringEncoding);
	}
    }
#elif defined(__MACOS8__)
    // No user name under MacOS 8.x
    return NULL;
#else
#error Dont know how to compute user name on this platform
#endif
    if (!__CFUserName)
        __CFUserName = CFRetain(CFSTR(""));
    return __CFUserName;
}

#define CFMaxHostNameLength	256
#define CFMaxHostNameSize	(CFMaxHostNameLength+1)

__private_extern__ CFStringRef _CFStringCreateHostName(void) {
#if defined(__MACOS8__)
        // Unsupported on Mac OS 8
        return NULL;
#else
    char myName[CFMaxHostNameSize];

    // return @"" instead of nil a la CFUserName() and Ali Ozer
    if (0 != gethostname(myName, CFMaxHostNameSize)) myName[0] = '\0';
    return CFStringCreateWithCString(NULL, myName, kCFPlatformInterfaceStringEncoding);
#endif
}

/* These are sanitized versions of the above functions. We might want to eliminate the above ones someday.
   These can return NULL.
*/
CF_EXPORT CFStringRef CFGetUserName(void) {
    return _CFUserName();
}

CF_EXPORT CFURLRef CFCopyHomeDirectoryURLForUser(CFStringRef uName) {
    return _CFCreateHomeDirectoryURLForUser(uName);
}

#undef PATH_LIST_SEP
#undef CFMaxHostNameLength
#undef CFMaxHostNameSize

