/*
 * Copyright (c) 2001 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*	CFRuntime.c
	Copyright 1999-2001, Apple, Inc. All rights reserved.
	Responsibility: Christopher Kane
*/

#include "CFRuntime.h"
#include "CFInternal.h"
#include <string.h>
#include <stdlib.h>
#if defined(__MACH__)
#include <monitor.h>
#endif

#if defined(__MACH__)
extern void __CFRecordAllocationEvent(int eventnum, void *ptr, int size, int data, const char *classname);
#else 
#define __CFRecordAllocationEvent(a, b, c, d, e)
#endif

enum {
// retain/release recording constants -- must match values
// used by OA for now; probably will change in the future
__kCFRetainEvent = 28,
__kCFReleaseEvent = 29
};


#if defined(__MACH__)

bool __CFOASafe = false;

void __CFOAInitialize() {
    static void (*dyfunc)(void) = (void *)0xFFFFFFFF;
    if (NULL == getenv("OAKeepAllocationStatistics")) return;
    if ((void *)0xFFFFFFFF == dyfunc) {
        dyfunc = NULL;
        if (NSIsSymbolNameDefined("__OAInitialize"))
            dyfunc = (void *)NSAddressOfSymbol(NSLookupAndBindSymbol("__OAInitialize"));
    }
    if (NULL != dyfunc) {
        dyfunc();
        __CFOASafe = true;
    }
}

void __CFRecordAllocationEvent(int eventnum, void *ptr, int size, int data, const char *classname) {
    static void (*dyfunc)(int, void *, int, int, const char *) = (void *)0xFFFFFFFF;
    if (!__CFOASafe) return;
    if ((void *)0xFFFFFFFF == dyfunc) {
        dyfunc = NULL;
        if (NSIsSymbolNameDefined("__OARecordAllocationEvent"))
            dyfunc = (void *)NSAddressOfSymbol(NSLookupAndBindSymbol("__OARecordAllocationEvent"));
    }
    if (NULL != dyfunc) {
        dyfunc(eventnum, ptr, size, data, classname);
    }
}

void __CFSetLastAllocationEventName(void *ptr, const char *classname) {
    static void (*dyfunc)(void *, const char *) = (void *)0xFFFFFFFF;
    if (!__CFOASafe) return;
    if ((void *)0xFFFFFFFF == dyfunc) {
        dyfunc = NULL;
        if (NSIsSymbolNameDefined("__OASetLastAllocationEventName"))
            dyfunc = (void *)NSAddressOfSymbol(NSLookupAndBindSymbol("__OASetLastAllocationEventName"));
    }
    if (NULL != dyfunc) {
        dyfunc(ptr, classname);
    }
}
#endif __MACH__


#define __kCFAllocatorTypeID_CONST	2

static CFTypeID __kCFNotATypeTypeID = _kCFRuntimeNotATypeID;

static const CFRuntimeClass __CFNotATypeClass = {
    0,
    "Not A Type",
    (void *)__CFAbort,
    (void *)__CFAbort,
    (void *)__CFAbort,
    (void *)__CFAbort,
    (void *)__CFAbort,
    (void *)__CFAbort,
    (void *)__CFAbort
};

static CFTypeID __kCFTypeTypeID = _kCFRuntimeNotATypeID;

static const CFRuntimeClass __CFTypeClass = {
    0,
    "CFType",
    (void *)__CFAbort,
    (void *)__CFAbort,
    (void *)__CFAbort,
    (void *)__CFAbort,
    (void *)__CFAbort,
    (void *)__CFAbort,
    (void *)__CFAbort
};

/* bits 15-8 in the CFRuntimeBase _info are type */
/* bits 7-0 in the CFRuntimeBase are reserved for CF's use */

static CFRuntimeClass * __CFRuntimeClassTable[256] = {NULL};
static SInt32 __CFRuntimeClassTableCount = 0;
static CFMutableDictionaryRef __CFRuntimeExternRefCountTable = NULL;

#if defined(__MACH__)

__private_extern__ SEL (*__CFGetObjCSelector)(const char *) = NULL;
__private_extern__ id (*__CFGetObjCClass)(const char *) = NULL;
__private_extern__ id (*__CFSendObjCMsg)(id, SEL, ...) = NULL;

#define __CFRUNTIMEOBJCTABLE_CNT 64
#define __CFRUNTIMEOBJCTABLE_SIZE (64 * sizeof(struct objc_class))

__private_extern__ struct objc_class *__CFRuntimeObjCClassTable = NULL;

__private_extern__ struct objc_class *__CFRuntimeObjCConstantStringClass = NULL;

__private_extern__ void __CFRuntimeTollFreeBridgeInitialize(void *default_class, int size) {
    CFIndex idx;
    if (NULL != default_class) {
	for (idx = 0; idx < __CFRUNTIMEOBJCTABLE_CNT; idx++) {
	    memmove(__CFRuntimeObjCClassTable + idx, default_class, size);
	}
    }
}

#endif

__private_extern__ void __CFRuntimeInitialize(void) {
#if defined(__MACH__)
    if (NULL == __CFRuntimeObjCClassTable) {
	// NOTE: this is not CFAllocatorAllocate() because allocators
	// aren't set up yet, and if we delay this until allocators
	// have been set up, then allocators don't have the correct
	// values for their isa pointers.
	__CFRuntimeObjCClassTable = malloc(__CFRUNTIMEOBJCTABLE_SIZE);
	memset(__CFRuntimeObjCClassTable, 0, __CFRUNTIMEOBJCTABLE_SIZE);
    }
#endif
    __kCFNotATypeTypeID = _CFRuntimeRegisterClass(&__CFNotATypeClass);
    __kCFTypeTypeID = _CFRuntimeRegisterClass(&__CFTypeClass);
}

__private_extern__ void __CFRuntimeInitialize2(void) {
    /*** _CFRuntimeCreateInstance() can finally be called after this line ***/
    // Creating this in CFRetain causes recursive call to CFRetain
    __CFRuntimeExternRefCountTable = CFDictionaryCreateMutable(kCFAllocatorSystemDefault, 0, NULL, NULL);
    // register another dummy type
    _CFRuntimeRegisterClass(&__CFNotATypeClass);
}

CFTypeID _CFRuntimeRegisterClass(const CFRuntimeClass * const cls) {
// version field must be 0
// className must be pure ASCII string, non-null
    __CFRuntimeClassTable[__CFRuntimeClassTableCount++] = (CFRuntimeClass *)cls;
    return __CFRuntimeClassTableCount - 1;
}

const CFRuntimeClass * const _CFRuntimeGetClassWithTypeID(CFTypeID typeID) {
    return __CFRuntimeClassTable[typeID];
}

CFTypeRef _CFRuntimeCreateInstance(CFAllocatorRef allocator, CFTypeID typeID, UInt32 extraBytes, unsigned char *category) {
    CFRuntimeBase *memory;
    Boolean usesSystemDefaultAllocator;
    SInt32 size;

    if (typeID <= __kCFTypeTypeID || __CFRuntimeClassTableCount <= typeID) {
	return NULL;
    }
    allocator = (NULL == allocator) ? __CFGetDefaultAllocator() : allocator;
    usesSystemDefaultAllocator = (allocator == kCFAllocatorSystemDefault);
    extraBytes = (extraBytes + (sizeof(void *) - 1)) & ~(sizeof(void *) - 1);
    size = sizeof(CFRuntimeBase) + extraBytes + (usesSystemDefaultAllocator ? 0 : sizeof(CFAllocatorRef));
    memory = CFAllocatorAllocate(allocator, size, 0);
    if (NULL == memory) {
	return NULL;
    }
#if defined(DEBUG) && !defined(__FREEBSD__)
    memset(memory, 0xCF, malloc_size(memory));
#endif
    if (__CFOASafe && category) {
        __CFSetLastAllocationEventName(memory, category);
    } else if (__CFOASafe) {
        __CFSetLastAllocationEventName(memory, __CFRuntimeClassTable[typeID]->className);
    }
    if (!usesSystemDefaultAllocator) {
	*(CFAllocatorRef *)((char *)memory) = CFRetain(allocator);
	memory = (CFRuntimeBase *)((char *)memory + sizeof(CFAllocatorRef));
    }
    memory->_isa = __CFISAForTypeID(typeID);
    memory->_rc = 0;
    memory->_info = 0;
    __CFBitfieldSetValue(memory->_info, 15, 8, typeID);
    if (usesSystemDefaultAllocator) {
	__CFBitfieldSetValue(memory->_info, 7, 7, 1);
    }
    if (NULL != __CFRuntimeClassTable[typeID]->init) {
	(__CFRuntimeClassTable[typeID]->init)(memory);
    }
    return memory;
}

__private_extern__ void __CFRuntimeSetInstanceISA(CFTypeRef cf, void *isa) {
    ((CFRuntimeBase *)cf)->_isa = isa;
}

void _CFRuntimeSetInstanceTypeID(CFTypeRef cf, CFTypeID typeID) {
    if (typeID <= __kCFTypeTypeID || __CFRuntimeClassTableCount <= typeID) {
	return;
    }
    __CFBitfieldSetValue(((CFRuntimeBase *)cf)->_info, 15, 8, typeID);
}

CF_INLINE CFTypeID __CFGenericTypeID(const void *cf) {
    return __CFBitfieldGetValue(((const CFRuntimeBase *)cf)->_info, 15, 8);
}

CFTypeID CFTypeGetTypeID() {
    return __kCFTypeTypeID;
}

__private_extern__ void __CFGenericValidateType_(CFTypeRef cf, CFTypeID type, char *func) {
    if (cf && CF_IS_OBJC(type, cf)) return;
    CFAssert2((cf != NULL) && (__CFGenericTypeID(cf) < __CFRuntimeClassTableCount) && (__kCFNotATypeTypeID != __CFGenericTypeID(cf)) && (__kCFTypeTypeID != __CFGenericTypeID(cf)), __kCFLogAssertion, "%s(): pointer 0x%x is not a CF object", func, cf); \
    CFAssert3(__CFGenericTypeID(cf) == type, __kCFLogAssertion, "%s(): pointer 0x%x is not a %s", func, cf, __CFRuntimeClassTable[type]->className);	\
}

#define __CFGenericAssertIsCF(cf) \
    CFAssert2(cf != NULL && __CFGenericTypeID(cf) < __CFRuntimeClassTableCount && __kCFNotATypeTypeID != __CFGenericTypeID(cf) && __kCFTypeTypeID != __CFGenericTypeID(cf), __kCFLogAssertion, "%s(): pointer 0x%x is not a CF object", __PRETTY_FUNCTION__, cf);

#if !defined(__MACH__)

#define CFTYPE_IS_OBJC(obj) (0)
#define CFTYPE_OBJC_FUNCDISPATCH0(rettype, obj, sel) do {} while (0)
#define CFTYPE_OBJC_FUNCDISPATCH1(rettype, obj, sel, a1) do {} while (0)

#endif !__MACH__

#if defined(__MACH__)

// Same as CF_IS_OBJC but returns NO for all CFTypes
#define CFTYPE_IS_OBJC(obj) \
	(((CFRuntimeBase *)(obj))->_isa != NULL && \
	(unsigned int)((CFRuntimeBase *)(obj))->_isa -	\
	(unsigned int)__CFRuntimeObjCClassTable >= __CFRUNTIMEOBJCTABLE_SIZE)

#define CFTYPE_OBJC_FUNCDISPATCH0(rettype, obj, sel) \
	if (__CFSendObjCMsg && CFTYPE_IS_OBJC(obj)) \
        {rettype (*func)(id, SEL) = (void *)__CFSendObjCMsg; \
        static SEL s = NULL; if (!s) s = __CFGetObjCSelector(sel); \
        if (s) return func((id)obj, s);}
#define CFTYPE_OBJC_FUNCDISPATCH1(rettype, obj, sel, a1) \
	if (__CFSendObjCMsg && CFTYPE_IS_OBJC(obj)) \
        {rettype (*func)(id, SEL, ...) = (void *)__CFSendObjCMsg; \
        static SEL s = NULL; if (!s) s = __CFGetObjCSelector(sel); \
        if (s) return func((id)obj, s, (a1));}

#endif __MACH__

CFTypeID CFGetTypeID(CFTypeRef cf) {
#if defined(DEBUG)
    if (NULL == cf) __CFAbort();
#endif
    CFTYPE_OBJC_FUNCDISPATCH0(CFTypeID, cf, "_cfTypeID");
    __CFGenericAssertIsCF(cf);
    return __CFGenericTypeID(cf);
}

CFStringRef CFCopyTypeIDDescription(CFTypeID type) {
    CFAssert2(type < __CFRuntimeClassTableCount && __kCFNotATypeTypeID != type && __kCFTypeTypeID != type, __kCFLogAssertion, "%s(): type %d is not a CF type ID", __PRETTY_FUNCTION__, type);
    return CFStringCreateWithCString(kCFAllocatorDefault, __CFRuntimeClassTable[type]->className, kCFStringEncodingASCII);
}

static UInt32 __CFGlobalRetainLock = 0;

 // Bit 31 (highest bit) in second word of cf instance indicates external ref count

#if 0

/* Returns non-zero if the extern ref count table needs incrementing */
// UInt32 __CFRetain(void *cf, UInt32 *rc);
__asm__ (
".text\n"
#if defined(__MACH__)
"_"
#endif
"__CFRetain:\n"
    "addi r3,r3,4\n"		// r3 = address of second word in cf instance
    "1: lwarx r5,0,r3\n"	// r5 = contents of second word
    "rlwinm r6,r5,16,17,31\n"	// r6 = inline ref count bits
    "addi r6,r6,1\n"		// increment r6
    "rlwimi r5,r6,16,1,15\n"	// replace ref count bits in r5 with new value in r6
    "rlwinm r7,r6,16,0,0\n"	// copy carry bit from r6 to high bit of r7 and zero rest
    "or r5,r5,r7\n"		// or extern ref count bit with carry bit
    "stwcx. r5,0,r3\n"		// try to store r5 to second word
    "bne- 1b\n"			// try again if reservation was lost
    "sync\n"			// should this be eieio or isync to be cheaper?
    "andi. r7,r6,0x7FFF\n"	// r7 = get new inline ref count bits
    "andi. r3,r6,0x8000\n"	// copy carry bit from r6 to r3
    "stw r7,0(r4)\n"		// return new inline ref count bits by reference
    "blr\n"
);

// faster version which is, unfortunately, not thread-safe in the presence of releases
CFTypeRef CFRetain(CFTypeRef cf) {
    UInt32 highBits, rc;
    if (NULL == cf) __CFAbort();
    CFTYPE_OBJC_FUNCDISPATCH0(CFTypeRef, cf, "retain");
    __CFGenericAssertIsCF(cf);
    if (__CFRetain(cf, &rc)) {
	UInt32 highBits;
	__CFSpinLock(&__CFGlobalRetainLock);
	// Roll over another bit to the external ref count; NULL return -> 0 ref count
	highBits = (CFIndex)CFDictionaryGetValue(__CFRuntimeExternRefCountTable, cf) + 1;
        CFDictionarySetValue(__CFRuntimeExternRefCountTable, cf, (void *)highBits);
	__CFSpinUnlock(&__CFGlobalRetainLock);
	// put low 31 bits of ref count in rc, 15 bits returned, 16 from highBits
	if (0xFFFF < highBits) {
	    rc = 0x7FFFFFFFUL;
	} else {
	    rc = ((highBits << 15) | rc);
	    if (rc < 0x7FFFFFFFFUL) rc++;
	}
    }
    if (__CFOASafe) __CFRecordAllocationEvent(__kCFRetainEvent, (void *)cf, 0, rc, NULL);
    return cf;
}
#endif 0

CFTypeRef CFRetain(CFTypeRef cf) {
    CFIndex highBits = 0, lowBits = 0;
    uint64_t compositeRC;
#if defined(DEBUG)
    if (NULL == cf) __CFAbort();
#endif
    CFTYPE_OBJC_FUNCDISPATCH0(CFTypeRef, cf, "retain");
    __CFGenericAssertIsCF(cf);
    __CFSpinLock(&__CFGlobalRetainLock);
    lowBits = ((CFRuntimeBase *)cf)->_rc;
    lowBits++;
    if ((lowBits & 0x07fff) == 0) {
	// Roll over another bit to the external ref count
	highBits = (CFIndex)CFDictionaryGetValue(__CFRuntimeExternRefCountTable, cf); // NULL return -> 0 ref count
	highBits++;
        CFDictionarySetValue(__CFRuntimeExternRefCountTable, cf, (void *)highBits);
	lowBits = 0x8000; // Bit 16 indicates external ref count
    }
    ((CFRuntimeBase *)cf)->_rc = lowBits;
    __CFSpinUnlock(&__CFGlobalRetainLock);
    compositeRC = (lowBits & 0x7fff) + (highBits << 15) + 1;
    if (compositeRC > 0x7fffffffULL) compositeRC = 0x7fffffffULL;
    if (__CFOASafe) __CFRecordAllocationEvent(__kCFRetainEvent, (void *)cf, 0, compositeRC, NULL);
    return cf;
}

#if defined(DEBUG)

/* CFZombieLevel levels:
 * 0xCF00FC00: unset yet
 *	0: free CF objects as normal
 *	1: scribble entire block with 0xFC and free
 *	2: same as 0
 *	3: scribble 0xFC, but not the header, and free
 *	16: don't scribble, and don't free
 *	17: scribble with 0xFC and don't free
 *	18: same as 4
 *	19: scribble 0xFC, but not the header, and don't free
 *
 * OR:
 *	bit 0: scribble deallocated CF objects with memory with 0xFC
 *	bit 1: do not scribble on CFRuntimeBase header (with bit 0)
 *	bit 4: do not free CF objects
 *	bit 7: use third-order byte as scribble byte (otherwise 0xFC)
 */

static uint32_t __CFZombieLevel = 0xCF00FC00;

static void __CFZombifyMemory(void *cf) {
#if defined (__MACH__)
    if (__CFZombieLevel & 0x1) 
	{
		void *ptr = cf;
		size_t size = malloc_size(cf);
		uint8_t byte = 0xFC;
		
		if (__CFZombieLevel & 0x2) 
		{
			ptr += sizeof(CFRuntimeBase);
			size -= sizeof(CFRuntimeBase);
		}
		
		if (__CFZombieLevel & 0x128) 
		{
			byte = (__CFZombieLevel >> 8) & 0xFF;
		}
		
		memset(ptr, byte, size);
    }
#endif
}

#endif /* DEBUG */

void CFRelease(CFTypeRef cf) {
    CFIndex highBits = 0, lowBits = 0;
    uint64_t compositeRC;
#if defined(DEBUG)
    if (NULL == cf) __CFAbort();
    if (0xCF00FC00 == __CFZombieLevel) {
        const char *value = getenv("CFZombieLevel");
        if (NULL != value) {
	    __CFZombieLevel = strtol(value, NULL, 0);
	}
    }
#endif
    CFTYPE_OBJC_FUNCDISPATCH0(void, cf, "release");
    __CFGenericAssertIsCF(cf);
    __CFSpinLock(&__CFGlobalRetainLock);
    lowBits = ((CFRuntimeBase *)cf)->_rc;
    if (0 == lowBits) {
        __CFSpinUnlock(&__CFGlobalRetainLock);
	if (__CFOASafe) __CFRecordAllocationEvent(__kCFReleaseEvent, (void *)cf, 0, 0, NULL);
	if (__kCFAllocatorTypeID_CONST == __CFGenericTypeID(cf)) {
#if defined(DEBUG)
	    __CFZombifyMemory(cf);
	    if (!(__CFZombieLevel & 0x16)) {
		__CFAllocatorDeallocate(cf);
	    }
#else
	    __CFAllocatorDeallocate(cf);
#endif
	} else {
	    CFAllocatorRef allocator;
	    if (NULL != __CFRuntimeClassTable[__CFGenericTypeID(cf)]->dealloc) {
		__CFRuntimeClassTable[__CFGenericTypeID(cf)]->dealloc(cf);
	    }
	    if (__CFBitfieldGetValue(((CFRuntimeBase *)cf)->_info, 7, 7)) {
		allocator = kCFAllocatorSystemDefault;
	    } else {
		allocator = CFGetAllocator(cf);
		cf -= sizeof(CFAllocatorRef);
	    }
#if defined(DEBUG)
	    __CFZombifyMemory(cf);
	    if (!(__CFZombieLevel & 0x16)) {
		CFAllocatorDeallocate(allocator, cf);
	    }
#else
	    CFAllocatorDeallocate(allocator, cf);
#endif
	    if (kCFAllocatorSystemDefault != allocator) {
		CFRelease(allocator);
	    }
	}
    } else {
	if (0x8000 == lowBits) {
	    // Time to remove a bit from the external ref count
            highBits = (CFIndex)CFDictionaryGetValue(__CFRuntimeExternRefCountTable, cf);	// NULL return -> 0 ref count
	    highBits--;
	    if (0 == highBits) {
		CFDictionaryRemoveValue(__CFRuntimeExternRefCountTable, cf);
		lowBits = 0x07fff;
	    } else {
           	CFDictionarySetValue(__CFRuntimeExternRefCountTable, cf, (void *)highBits);
		lowBits = 0x0ffff;
	    }
 	} else {
	    lowBits--;
	}
	((CFRuntimeBase *)cf)->_rc = lowBits;
        __CFSpinUnlock(&__CFGlobalRetainLock);
	compositeRC = (lowBits & 0x7fff) + (highBits << 15) + 1;
	if (compositeRC > 0x7fffffffULL) compositeRC = 0x7fffffffULL;
	if (__CFOASafe) __CFRecordAllocationEvent(__kCFReleaseEvent, (void *)cf, 0, compositeRC, NULL);
    }
}

static uint64_t __CFGetFullRetainCount(CFTypeRef cf) {
    UInt32 lowBits = 0;
    uint64_t highBits = 0, compositeRC;
    lowBits = ((CFRuntimeBase *)cf)->_rc;
    if ((lowBits & 0x08000) != 0) {
        highBits = (UInt32)CFDictionaryGetValue(__CFRuntimeExternRefCountTable, cf);
    }
    compositeRC = (lowBits & 0x7fff) + (highBits << 15) + 1;
    return compositeRC;
}

CFIndex CFGetRetainCount(CFTypeRef cf) {
    uint64_t rc;
    CFIndex result;
#if defined(DEBUG)
    if (NULL == cf) __CFAbort();
#endif
    CFTYPE_OBJC_FUNCDISPATCH0(CFIndex, cf, "retainCount");
    __CFGenericAssertIsCF(cf);
    rc = __CFGetFullRetainCount(cf);
    result = (rc < 0x7FFFFFFFFUL) ? (CFIndex)rc : (CFIndex)0x7FFFFFFFFL;
    return result;
}

Boolean CFEqual(CFTypeRef cf1, CFTypeRef cf2) {
#if defined(DEBUG)
    if (NULL == cf1) __CFAbort();
    if (NULL == cf2) __CFAbort();
#endif
    if (cf1 == cf2) return true;
    CFTYPE_OBJC_FUNCDISPATCH1(Boolean, cf1, "isEqual:", cf2);
    CFTYPE_OBJC_FUNCDISPATCH1(Boolean, cf2, "isEqual:", cf1);
    __CFGenericAssertIsCF(cf1);
    __CFGenericAssertIsCF(cf2);
    if (__CFGenericTypeID(cf1) != __CFGenericTypeID(cf2)) return false;
    if (NULL != __CFRuntimeClassTable[__CFGenericTypeID(cf1)]->equal) {
	return __CFRuntimeClassTable[__CFGenericTypeID(cf1)]->equal(cf1, cf2);
    }
    return false;
}

CFHashCode CFHash(CFTypeRef cf) {
#if defined(DEBUG)
    if (NULL == cf) __CFAbort();
#endif
    CFTYPE_OBJC_FUNCDISPATCH0(CFHashCode, cf, "hash");
    __CFGenericAssertIsCF(cf);
    if (NULL != __CFRuntimeClassTable[__CFGenericTypeID(cf)]->hash) {
	return __CFRuntimeClassTable[__CFGenericTypeID(cf)]->hash(cf);
    }
    return (CFHashCode)cf;
}

// definition: produces a normally non-NULL debugging description of the object
CFStringRef CFCopyDescription(CFTypeRef cf) {
#if defined(DEBUG)
    if (NULL == cf) __CFAbort();
#endif
    CFTYPE_OBJC_FUNCDISPATCH0(CFStringRef, cf, "_copyDescription");
    __CFGenericAssertIsCF(cf);
    if (NULL != __CFRuntimeClassTable[__CFGenericTypeID(cf)]->copyDebugDesc) {
	CFStringRef result;
	result = __CFRuntimeClassTable[__CFGenericTypeID(cf)]->copyDebugDesc(cf);
	if (NULL != result) return result;
    }
    return CFStringCreateWithFormat(kCFAllocatorDefault, NULL, CFSTR("<%s %p [%p]>"), __CFRuntimeClassTable[__CFGenericTypeID(cf)]->className, cf, CFGetAllocator(cf));
}

// Definition: if type produces a formatting description, return that string, otherwise NULL
__private_extern__ CFStringRef __CFCopyFormattingDescription(CFTypeRef cf, CFDictionaryRef formatOptions) {
#if defined(DEBUG)
    if (NULL == cf) __CFAbort();
#endif
#if defined(__MACH__)
    if (__CFSendObjCMsg && CFTYPE_IS_OBJC(cf)) {
	static SEL s = NULL, r = NULL;
	CFStringRef (*func)(id, SEL, ...) = (void *)__CFSendObjCMsg;
	if (!s) s = __CFGetObjCSelector("_copyFormattingDescription:");
	if (!r) r = __CFGetObjCSelector("respondsToSelector:");
	if (s && func((id)cf, r, s)) return func((id)cf, s, formatOptions);
	return NULL;
    }
#endif __MACH__
    __CFGenericAssertIsCF(cf);
    if (NULL != __CFRuntimeClassTable[__CFGenericTypeID(cf)]->copyFormattingDesc) {
	return __CFRuntimeClassTable[__CFGenericTypeID(cf)]->copyFormattingDesc(cf, formatOptions);
    }
    return NULL;
}

CFAllocatorRef CFGetAllocator(CFTypeRef cf) {
#if defined(DEBUG)
    if (NULL == cf) __CFAbort();
#endif
// CF: need to get allocator from objc objects in better way...how?
// -> bridging of CFAllocators and malloc_zone_t will help this
    if (CFTYPE_IS_OBJC(cf)) return __CFGetDefaultAllocator();
    __CFGenericAssertIsCF(cf);
    if (__kCFAllocatorTypeID_CONST == __CFGenericTypeID(cf)) {
	extern CFAllocatorRef __CFAllocatorGetAllocator(CFTypeRef cf);
	return __CFAllocatorGetAllocator(cf);
    }
    if (__CFBitfieldGetValue(((CFRuntimeBase *)cf)->_info, 7, 7)) {
	return kCFAllocatorSystemDefault;
    }
    return *(CFAllocatorRef *)((char *)cf - sizeof(CFAllocatorRef));
}

extern void __CFBaseInitialize(void);
extern void __CFRuntimeInitialize(void);
extern void __CFRuntimeInitialize2(void);
extern void __CFAllocatorInitialize(void);
extern void __CFStringInitialize(void);
extern void __CFArrayInitialize(void);
extern void __CFBagInitialize(void);
extern void __CFBooleanInitialize(void);
extern void __CFCharacterSetInitialize(void);
extern void __CFDataInitialize(void);
extern void __CFDateInitialize(void);
extern void __CFDictionaryInitialize(void);
extern void __CFNumberInitialize(void);
extern void __CFSetInitialize(void);
extern void __CFStorageInitialize(void);
extern void __CFTimeZoneInitialize(void);
extern void __CFTreeInitialize(void);
extern void __CFURLInitialize(void);
extern void __CFUtilitiesInitialize(void);	// not a CF type
extern void __CFXMLNodeInitialize(void);
extern void __CFXMLParserInitialize(void);
#if defined(__MACH__)
extern void __CFMessagePortInitialize(void);
extern void __CFMachPortInitialize(void);
extern void __CFRunLoopInitialize(void);
extern void __CFRunLoopObserverInitialize(void);
extern void __CFRunLoopSourceInitialize(void);
extern void __CFRunLoopTimerInitialize(void);
extern void __CFSocketInitialize(void);
#endif
extern void __CFBundleInitialize(void);
extern void __CFPlugInInitialize(void);
extern void __CFPlugInInstanceInitialize(void);
extern void __CFUUIDInitialize(void);
extern void __CFURLAccessInitialize(void); // not a CF type
#if !defined(__MACOS8__)
extern void __CFBinaryHeapInitialize(void);
extern void __CFBitVectorInitialize(void);
#endif 

#define DO_SYSCALL_TRACE_HELPERS 1
#if defined(DO_SYSCALL_TRACE_HELPERS) && defined(__MACH__)
extern void ptrace(int, int, int, int);
#define SYSCALL_TRACE(N)        do ptrace(N, 0, 0, 0); while (0)
#else
#define SYSCALL_TRACE(N)        do {} while (0)
#endif

#if defined(__MACH__) && defined(PROFILE)
static void _CF_mcleanup(void) {
    monitor(0,0,0,0,0);
}
#endif

#if defined(__LINUX__) || defined(__FREEBSD__)
static void __CFInitialize(void) __attribute__ ((constructor));
static
#endif
void __CFInitialize(void) {
    static int __done = 0;
    if (!__done) {
	__done = 1;
	SYSCALL_TRACE(0xC000);

#if defined(__MACH__) && defined(PROFILE)
	{
	    const char *v = getenv("DYLD_IMAGE_SUFFIX");
	    if (v && 0 == strcmp("_profile", v)) {
		// Unfortunately, no way to know if this has already been done,
		// or I would not do it.  Not much information will be lost.
		atexit(_CF_mcleanup);
		moninit();
	    }
	}
#endif

	__CFBaseInitialize();

	/* Here so that two runtime classes get indices 0, 1. */
	__CFRuntimeInitialize();

	/* Here so that __kCFAllocatorTypeID gets index 2. */
	__CFAllocatorInitialize();

	/* Basic collections need to be up before CFString. */
	__CFDictionaryInitialize();
	__CFArrayInitialize();
	__CFDataInitialize();
	__CFRuntimeInitialize2();
	__CFStringInitialize();
	__CFUtilitiesInitialize();

	__CFBagInitialize();
	__CFBooleanInitialize();
	__CFCharacterSetInitialize();
	__CFDateInitialize();
	__CFNumberInitialize();
	__CFSetInitialize();
	__CFStorageInitialize();
	__CFTimeZoneInitialize();
	__CFTreeInitialize();
	__CFURLInitialize();
        __CFXMLNodeInitialize();
	__CFXMLParserInitialize();
	__CFBundleInitialize();
	__CFPlugInInitialize();
	__CFPlugInInstanceInitialize();
	__CFUUIDInitialize();
	__CFURLAccessInitialize(); // not a CF type
#if defined(__MACH__)
	__CFMessagePortInitialize();
	__CFMachPortInitialize();
	__CFRunLoopInitialize();
	__CFRunLoopObserverInitialize();
	__CFRunLoopSourceInitialize();
	__CFRunLoopTimerInitialize();
	__CFSocketInitialize();
#endif
#if !defined(__MACOS8__)
	__CFBinaryHeapInitialize();
	__CFBitVectorInitialize();
#endif

	SYSCALL_TRACE(0xC001);
#if defined(__MACH__)
        __CFOAInitialize();
	SYSCALL_TRACE(0xC003);
#endif
#if defined(DEBUG) && !defined(__MACOS8__)
	// Don't log on MacOS 8 as this will create a log file unnecessarily
	CFLog (0, CFSTR("Assertions enabled"));
#endif
	SYSCALL_TRACE(0xC0FF);
    }
}


#undef DO_SYSCALL_TRACE_HELPERS
#undef SYSCALL_TRACE
#undef __kCFAllocatorTypeID_CONST
#undef __CFGenericAssertIsCF

