/*
 * Copyright (c) 2001 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*	CFDictionary.c
	Copyright 1998-2001, Apple, Inc. All rights reserved.
	Responsibility: Christopher Kane
*/

#include <CoreFoundation/CFDictionary.h>
#include "CFVeryPrivate.h"
#include "CFInternal.h"

#if defined(__MACH__) && defined(__ppc__)
// Nasty performance hack: we mark these callbacks as "faulted" on PPC
const CFDictionaryKeyCallBacks kCFTypeDictionaryKeyCallBacks = {0, (void *)((unsigned)__CFTypeCollectionRetain+1), (void *)((unsigned)__CFTypeCollectionRelease+1), (void *)((unsigned)CFCopyDescription+1), (void *)((unsigned)CFEqual+1), (void *)((unsigned)CFHash+1)};
const CFDictionaryKeyCallBacks kCFCopyStringDictionaryKeyCallBacks = {0, (void *)((unsigned)CFStringCreateCopy+1), (void *)((unsigned)__CFTypeCollectionRelease+1), (void *)((unsigned)CFCopyDescription+1), (void *)((unsigned)CFEqual+1), (void *)((unsigned)CFHash+1)};
const CFDictionaryValueCallBacks kCFTypeDictionaryValueCallBacks = {0, (void *)((unsigned)__CFTypeCollectionRetain+1), (void *)((unsigned)__CFTypeCollectionRelease+1), (void *)((unsigned)CFCopyDescription+1), (void *)((unsigned)CFEqual+1)};
#else
const CFDictionaryKeyCallBacks kCFTypeDictionaryKeyCallBacks = {0, __CFTypeCollectionRetain, __CFTypeCollectionRelease, CFCopyDescription, CFEqual, CFHash};
const CFDictionaryKeyCallBacks kCFCopyStringDictionaryKeyCallBacks = {0, (void *)CFStringCreateCopy, __CFTypeCollectionRelease, CFCopyDescription, CFEqual, CFHash};
const CFDictionaryValueCallBacks kCFTypeDictionaryValueCallBacks = {0, __CFTypeCollectionRetain, __CFTypeCollectionRelease, CFCopyDescription, CFEqual};
#endif
static const CFDictionaryKeyCallBacks __kCFNullDictionaryKeyCallBacks = {0, NULL, NULL, NULL, NULL, NULL};
static const CFDictionaryValueCallBacks __kCFNullDictionaryValueCallBacks = {0, NULL, NULL, NULL, NULL};

/* Hashing size tables.  The number of buckets is a power of two:
        #buckets(x) = 4 * (1 << x)
The corresponding maximum load or capacity is given by:
        maxcap(x) = floor((#buckets(x) * 2.0 + 2.0) / 3.0)
or approximately 67%.  'x' ranges from 0 to 29.
*/

static const CFIndex __CFDictionaryCapacities[30] = {
    3, 6, 11, 22, 43, 86, 171, 342, 683, 1366, 2731, 5462, 10923,
    21846, 43691, 87382, 174763, 349526, 699051, 1398102, 2796203,
    5592406, 11184811, 22369622, 44739243, 89478486, 178956971,
    357913942, 715827883, 1431655766
};

CF_INLINE CFIndex __CFDictionaryRoundUpCapacity(CFIndex capacity) {
    CFIndex idx;
    for (idx = 0; idx < 30 && __CFDictionaryCapacities[idx] < capacity; idx++);
    return __CFDictionaryCapacities[idx];
}

CF_INLINE CFIndex __CFDictionaryNumBucketsForCapacity(CFIndex capacity) {
    CFIndex idx;
    for (idx = 0; idx < 30 && __CFDictionaryCapacities[idx] < capacity; idx++);
    return 4 * (1 << idx);
}

enum {		/* Bits 1-0 */
    __kCFDictionaryImmutable = 0,	/* unchangable and fixed capacity */
    __kCFDictionaryMutable = 1,		/* changeable and variable capacity */
    __kCFDictionaryFixedMutable = 3	/* changeable and fixed capacity */
};

enum {		/* Bits 5-4 (value), 3-2 (key) */
    __kCFDictionaryHasNullCallBacks = 0,
    __kCFDictionaryHasCFTypeCallBacks = 1,
    __kCFDictionaryHasCustomCallBacks = 3	/* callbacks are at end of header */
};

struct __CFDictionaryBucket {
    const void *_key;
};

struct __CFDictionary {
    CFRuntimeBase _base;
    CFIndex _count;		/* number of values */
    CFIndex _capacity;		/* maximum number of values */
    CFIndex _bucketsNum;	/* number of slots */
    const void *_emptyMarker;
    const void *_deletedMarker;
    void *_context;		/* private */
    struct __CFDictionaryBucket *_buckets;	/* can be NULL if not allocated yet */
    const void **_values;			/* can be NULL if not allocated yet */
};

CF_INLINE Boolean __CFDictionaryBucketIsEmpty(CFDictionaryRef dict, const struct __CFDictionaryBucket *bucket) {
    return (dict->_emptyMarker == bucket->_key);
}

CF_INLINE Boolean __CFDictionaryBucketIsDeleted(CFDictionaryRef dict, const struct __CFDictionaryBucket *bucket) {
    return (dict->_deletedMarker == bucket->_key);
}

CF_INLINE Boolean __CFDictionaryBucketIsOccupied(CFDictionaryRef dict, const struct __CFDictionaryBucket *bucket) {
    return (dict->_emptyMarker != bucket->_key && dict->_deletedMarker != bucket->_key);
}

/* Bits 1-0 of the base reserved bits are used for mutability variety */
/* Bits 3-2 of the base reserved bits are used for key callback indicator bits */
/* Bits 5-4 of the base reserved bits are used for value callback indicator bits */

CF_INLINE CFIndex __CFDictionaryGetType(CFDictionaryRef dict) {
    return __CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 1, 0);
}

CF_INLINE CFIndex __CFDictionaryGetSizeOfType(CFIndex t) {
    CFIndex size = sizeof(struct __CFDictionary);
    if (__CFBitfieldGetValue(t, 3, 2) == __kCFDictionaryHasCustomCallBacks) {
	size += sizeof(CFDictionaryKeyCallBacks);
    }
    if (__CFBitfieldGetValue(t, 5, 4) == __kCFDictionaryHasCustomCallBacks) {
	size += sizeof(CFDictionaryValueCallBacks);
    }
    return size;
}

CF_INLINE const CFDictionaryKeyCallBacks *__CFDictionaryGetKeyCallBacks(CFDictionaryRef dict) {
    CFDictionaryKeyCallBacks *result = NULL;
    switch (__CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 3, 2)) {
    case __kCFDictionaryHasNullCallBacks:
	return &__kCFNullDictionaryKeyCallBacks;
    case __kCFDictionaryHasCFTypeCallBacks:
	return &kCFTypeDictionaryKeyCallBacks;
    case __kCFDictionaryHasCustomCallBacks:
	break;
    }
    result = (CFDictionaryKeyCallBacks *)((uint8_t *)dict + sizeof(struct __CFDictionary));
    return result;
}

CF_INLINE Boolean __CFDictionaryKeyCallBacksMatchNull(const CFDictionaryKeyCallBacks *c) {
    return (NULL == c ||
	(c->retain == __kCFNullDictionaryKeyCallBacks.retain &&
	 c->release == __kCFNullDictionaryKeyCallBacks.release &&
	 c->copyDescription == __kCFNullDictionaryKeyCallBacks.copyDescription &&
	 c->equal == __kCFNullDictionaryKeyCallBacks.equal &&
	 c->hash == __kCFNullDictionaryKeyCallBacks.hash));
}

CF_INLINE Boolean __CFDictionaryKeyCallBacksMatchCFType(const CFDictionaryKeyCallBacks *c) {
    return (&kCFTypeDictionaryKeyCallBacks == c ||
	(c->retain == kCFTypeDictionaryKeyCallBacks.retain &&
	 c->release == kCFTypeDictionaryKeyCallBacks.release &&
	 c->copyDescription == kCFTypeDictionaryKeyCallBacks.copyDescription &&
	 c->equal == kCFTypeDictionaryKeyCallBacks.equal &&
	 c->hash == kCFTypeDictionaryKeyCallBacks.hash));
}

CF_INLINE const CFDictionaryValueCallBacks *__CFDictionaryGetValueCallBacks(CFDictionaryRef dict) {
    CFDictionaryValueCallBacks *result = NULL;
    switch (__CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 5, 4)) {
    case __kCFDictionaryHasNullCallBacks:
	return &__kCFNullDictionaryValueCallBacks;
    case __kCFDictionaryHasCFTypeCallBacks:
	return &kCFTypeDictionaryValueCallBacks;
    case __kCFDictionaryHasCustomCallBacks:
	break;
    }
    if (__CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 3, 2) == __kCFDictionaryHasCustomCallBacks) {
	result = (CFDictionaryValueCallBacks *)((uint8_t *)dict + sizeof(struct __CFDictionary) + sizeof(CFDictionaryKeyCallBacks));
    } else {
	result = (CFDictionaryValueCallBacks *)((uint8_t *)dict + sizeof(struct __CFDictionary));
    }
    return result;
}

CF_INLINE Boolean __CFDictionaryValueCallBacksMatchNull(const CFDictionaryValueCallBacks *c) {
    return (NULL == c ||
	(c->retain == __kCFNullDictionaryValueCallBacks.retain &&
	 c->release == __kCFNullDictionaryValueCallBacks.release &&
	 c->copyDescription == __kCFNullDictionaryValueCallBacks.copyDescription &&
	 c->equal == __kCFNullDictionaryValueCallBacks.equal));
}

CF_INLINE Boolean __CFDictionaryValueCallBacksMatchCFType(const CFDictionaryValueCallBacks *c) {
    return (&kCFTypeDictionaryValueCallBacks == c ||
	(c->retain == kCFTypeDictionaryValueCallBacks.retain &&
	 c->release == kCFTypeDictionaryValueCallBacks.release &&
	 c->copyDescription == kCFTypeDictionaryValueCallBacks.copyDescription &&
	 c->equal == kCFTypeDictionaryValueCallBacks.equal));
}


/* Hashing theory: double hashing is used, with the hash function being:
	h(keyhash, i) = (h1(keyhash) + i*h2(keyhash)) mod #buckets
where h1 and h2 are secondary hash functions.  h1 is used to compute the
starting probe position and is defined to be:
	h1(keyhash) = keyhash mod #buckets
h2 is used as the skip or increment value as probing progresses:
	h2(keyhash) = (2 * keyhash + 1) mod #buckets
Since h2 also depends on the hash of the key, many more probe sequences
are possible than with, say, linear probing.  The number of buckets in
the tables is always a power of two and h2 is always odd to ensure that
the value of h2 is always relatively prime to the size of the table.
Other h2() functions (as long as the value is always odd) can be chosen.

Note that many hash values have only low-order bits set, with upper bits
all zero.  So trying to use some of the upper bits, which are thrown away
by the modulus operations, will often result in very uninteresting h2().

Another way to ensure relatively primality would be to define h2 as:
	h2(keyhash) = 1 + (keyhash mod (#buckets - 1))
and make the number of buckets in the tables always prime.  But the
former method is chosen here, because it is somewhat faster.

Other possible optimizations:
- On each search, swap the found bucket contents with the starting bucket's contents
- Better h2()
- Rehash after #buckets/4 removes to cut down on number of deleted entries

*/

static void __CFDictionaryFindBuckets1a(CFDictionaryRef dict, const void *key, CFIndex *match) {
    struct __CFDictionaryBucket *buckets = dict->_buckets;
    CFHashCode keyHash = (CFHashCode)key;
    UInt32 mask = dict->_bucketsNum - 1;
    CFIndex probe = keyHash & mask;
    CFIndex probeskip = ((keyHash << 1) + 1) & mask;
    CFIndex start = probe;
    *match = kCFNotFound;
    for (;;) {
	struct __CFDictionaryBucket *currentBucket = buckets + probe;
	if (__CFDictionaryBucketIsEmpty(dict, currentBucket)) {
	    return;
	} else if (__CFDictionaryBucketIsDeleted(dict, currentBucket)) {
	    /* do nothing */
	} else if (currentBucket->_key == key) {
	    *match = probe;
	    return;
	}
	probe = (probe + probeskip) & mask;
	if (start == probe) return;
    }
}

static void __CFDictionaryFindBuckets1b(CFDictionaryRef dict, const void *key, CFIndex *match) {
    const CFDictionaryKeyCallBacks *cb = __CFDictionaryGetKeyCallBacks(dict);
    struct __CFDictionaryBucket *buckets = dict->_buckets;
    CFHashCode keyHash = cb->hash ? (CFHashCode)INVOKE_CALLBACK2(cb->hash, key, dict->_context) : (CFHashCode)key;
    UInt32 mask = dict->_bucketsNum - 1;
    CFIndex probe = keyHash & mask;
    CFIndex probeskip = ((keyHash << 1) + 1) & mask;
    CFIndex start = probe;
    *match = kCFNotFound;
    for (;;) {
	struct __CFDictionaryBucket *currentBucket = buckets + probe;
	if (__CFDictionaryBucketIsEmpty(dict, currentBucket)) {
	    return;
	} else if (__CFDictionaryBucketIsDeleted(dict, currentBucket)) {
	    /* do nothing */
	} else if (currentBucket->_key == key || (cb->equal && INVOKE_CALLBACK3(cb->equal, currentBucket->_key, key, dict->_context))) {
	    *match = probe;
	    return;
	}
	probe = (probe + probeskip) & mask;
	if (start == probe) return;
    }
}

static void __CFDictionaryFindBuckets2(CFDictionaryRef dict, const void *key, CFIndex *match, CFIndex *nomatch) {
    const CFDictionaryKeyCallBacks *cb = __CFDictionaryGetKeyCallBacks(dict);
    struct __CFDictionaryBucket *buckets = dict->_buckets;
    CFHashCode keyHash = cb->hash ? (CFHashCode)INVOKE_CALLBACK2(cb->hash, key, dict->_context) : (CFHashCode)key;
    UInt32 mask = dict->_bucketsNum - 1;
    CFIndex probe = keyHash & mask;
    CFIndex probeskip = ((keyHash << 1) + 1) & mask;
    CFIndex start = probe;
    *match = kCFNotFound;
    *nomatch = kCFNotFound;
    for (;;) {
	struct __CFDictionaryBucket *currentBucket = buckets + probe;
	if (__CFDictionaryBucketIsEmpty(dict, currentBucket)) {
	    if (kCFNotFound == *nomatch) *nomatch = probe;
	    return;
	} else if (__CFDictionaryBucketIsDeleted(dict, currentBucket)) {
	    if (kCFNotFound == *nomatch) *nomatch = probe;
	    if (kCFNotFound != *match) return;
	} else if (kCFNotFound == *match && (currentBucket->_key == key || (cb->equal && INVOKE_CALLBACK3(cb->equal, currentBucket->_key, key, dict->_context)))) {
	    *match = probe;
	    if (kCFNotFound != *nomatch) return;
	}
	probe = (probe + probeskip) & mask;
	if (start == probe) return;
    }
}

static void __CFDictionaryFindNewEmptyMarker(CFDictionaryRef dict) {
    struct __CFDictionaryBucket *buckets;
    const void *newEmpty;
    Boolean hit;
    CFIndex idx, nbuckets;
    buckets = dict->_buckets;
    nbuckets = dict->_bucketsNum;
    newEmpty = dict->_emptyMarker;
    do {
	newEmpty -= 2;
	hit = false;
	for (idx = 0; idx < nbuckets; idx++) {
	    if (newEmpty == buckets[idx]._key) {
		hit = true;
		break;
	    }
	}
    } while (hit);
    for (idx = 0; idx < nbuckets; idx++) {
	if (dict->_emptyMarker == buckets[idx]._key) {
	    buckets[idx]._key = newEmpty;
	}
    }
    ((struct __CFDictionary *)dict)->_emptyMarker = newEmpty;
}

static void __CFDictionaryFindNewDeletedMarker(CFDictionaryRef dict) {
    struct __CFDictionaryBucket *buckets;
    const void *newDeleted;
    Boolean hit;
    CFIndex idx, nbuckets;
    buckets = dict->_buckets;
    nbuckets = dict->_bucketsNum;
    newDeleted = dict->_deletedMarker;
    do {
	newDeleted += 2;
	hit = false;
	for (idx = 0; idx < nbuckets; idx++) {
	    if (newDeleted == buckets[idx]._key) {
		hit = true;
		break;
	    }
	}
    } while (hit);
    for (idx = 0; idx < nbuckets; idx++) {
	if (dict->_deletedMarker == buckets[idx]._key) {
	    buckets[idx]._key = newDeleted;
	}
    }
    ((struct __CFDictionary *)dict)->_deletedMarker = newDeleted;
}

static Boolean __CFDictionaryEqual(CFTypeRef cf1, CFTypeRef cf2) {
    CFDictionaryRef dict1 = (CFDictionaryRef)cf1;
    CFDictionaryRef dict2 = (CFDictionaryRef)cf2;
    const CFDictionaryKeyCallBacks *cb1, *cb2;
    const CFDictionaryValueCallBacks *vcb1, *vcb2;
    const struct __CFDictionaryBucket *buckets;
    CFIndex idx, nbuckets;
    if (dict1 == dict2) return true;
    if (dict1->_count != dict2->_count) return false;
    cb1 = __CFDictionaryGetKeyCallBacks(dict1);
    cb2 = __CFDictionaryGetKeyCallBacks(dict2);
    if (cb1->equal != cb2->equal) return false;
    vcb1 = __CFDictionaryGetValueCallBacks(dict1);
    vcb2 = __CFDictionaryGetValueCallBacks(dict2);
    if (vcb1->equal != vcb2->equal) return false;
    if (0 == dict1->_count) return true; /* after function comparison! */
    buckets = dict1->_buckets;
    nbuckets = dict1->_bucketsNum;
    for (idx = 0; idx < nbuckets; idx++) {
	if (__CFDictionaryBucketIsOccupied(dict1, &buckets[idx])) {
	    const void *value;
	    if (!CFDictionaryGetValueIfPresent(dict2, buckets[idx]._key, &value)) return false;
	    if (dict1->_values[idx] != value && vcb1->equal && !INVOKE_CALLBACK3(vcb1->equal, dict1->_values[idx], value, dict1->_context)) {
		return false;
	    }
	}
    }
    return true;
}

static CFHashCode __CFDictionaryHash(CFTypeRef cf) {
    CFDictionaryRef dict = (CFDictionaryRef)cf;
    return dict->_count;
}

static CFStringRef __CFDictionaryCopyDescription(CFTypeRef cf) {
    CFDictionaryRef dict = (CFDictionaryRef)cf;
    const CFDictionaryKeyCallBacks *cb;
    const CFDictionaryValueCallBacks *vcb;
    const struct __CFDictionaryBucket *buckets;
    CFIndex idx, nbuckets;
    CFMutableStringRef result;
    cb = __CFDictionaryGetKeyCallBacks(dict);
    vcb = __CFDictionaryGetValueCallBacks(dict);
    buckets = dict->_buckets;
    nbuckets = dict->_bucketsNum;
    result = CFStringCreateMutable(kCFAllocatorSystemDefault, 0);
    CFStringAppendFormat(result, NULL, CFSTR("<CFDictionary %p [%p]>{count = %u, capacity = %u, pairs = (\n"), dict, CFGetAllocator(dict), dict->_count, dict->_capacity);
    for (idx = 0; idx < nbuckets; idx++) {
	if (__CFDictionaryBucketIsOccupied(dict, &buckets[idx])) {
	    CFStringRef kDesc = NULL, vDesc = NULL;
	    if (NULL != cb->copyDescription) {
		kDesc = (CFStringRef)INVOKE_CALLBACK2(cb->copyDescription, buckets[idx]._key, dict->_context);
	    }
	    if (NULL != vcb->copyDescription) {
		vDesc = (CFStringRef)INVOKE_CALLBACK2(vcb->copyDescription, dict->_values[idx], dict->_context);
	    }
	    if (NULL != kDesc && NULL != vDesc) {
		CFStringAppendFormat(result, NULL, CFSTR("\t%u : %@ = %@\n"), idx, kDesc, vDesc);
		CFRelease(kDesc);
		CFRelease(vDesc);
	    } else if (NULL != kDesc) {
		CFStringAppendFormat(result, NULL, CFSTR("\t%u : %@ = <%p>\n"), idx, kDesc, dict->_values[idx]);
		CFRelease(kDesc);
	    } else if (NULL != vDesc) {
		CFStringAppendFormat(result, NULL, CFSTR("\t%u : <%p> = %@\n"), idx, buckets[idx]._key, vDesc);
		CFRelease(vDesc);
	    } else {
		CFStringAppendFormat(result, NULL, CFSTR("\t%u : <%p> = <%p>\n"), idx, buckets[idx]._key, dict->_values[idx]);
	    }
	}
    }
    CFStringAppend(result, CFSTR(")}"));
    return result;
}

static void __CFDictionaryDeallocate(CFTypeRef cf) {
    CFMutableDictionaryRef dict = (CFMutableDictionaryRef)cf;
    CFAllocatorRef allocator = CFGetAllocator(dict);
    if (__CFDictionaryGetType(dict) == __kCFDictionaryImmutable) {
        __CFBitfieldSetValue(((CFRuntimeBase *)dict)->_info, 1, 0, __kCFDictionaryFixedMutable);
    }
    CFDictionaryRemoveAllValues(dict);
    if (__CFDictionaryGetType(dict) == __kCFDictionaryMutable && dict->_buckets) {
	CFAllocatorDeallocate(allocator, dict->_buckets);
    }
}

static CFTypeID __kCFDictionaryTypeID = _kCFRuntimeNotATypeID;

static const CFRuntimeClass __CFDictionaryClass = {
    0,
    "CFDictionary",
    NULL,	// init
    NULL,	// copy
    __CFDictionaryDeallocate,
    __CFDictionaryEqual,
    __CFDictionaryHash,
    NULL,	// 
    __CFDictionaryCopyDescription
};

__private_extern__ void __CFDictionaryInitialize(void) {
    __kCFDictionaryTypeID = _CFRuntimeRegisterClass(&__CFDictionaryClass);
}

CFTypeID CFDictionaryGetTypeID(void) {
    return __kCFDictionaryTypeID;
}

static CFDictionaryRef __CFDictionaryInit(CFAllocatorRef allocator, UInt32 flags, CFIndex capacity, const CFDictionaryKeyCallBacks *callBacks, const CFDictionaryValueCallBacks *valueCallBacks) {
    struct __CFDictionary *memory;
    UInt32 size;
    CFIndex idx;
    __CFBitfieldSetValue(flags, 31, 2, 0);
    if (__CFDictionaryKeyCallBacksMatchNull(callBacks)) {
	__CFBitfieldSetValue(flags, 3, 2, __kCFDictionaryHasNullCallBacks);
    } else if (__CFDictionaryKeyCallBacksMatchCFType(callBacks)) {
	__CFBitfieldSetValue(flags, 3, 2, __kCFDictionaryHasCFTypeCallBacks);
    } else {
	__CFBitfieldSetValue(flags, 3, 2, __kCFDictionaryHasCustomCallBacks);
    }
    if (__CFDictionaryValueCallBacksMatchNull(valueCallBacks)) {
	__CFBitfieldSetValue(flags, 5, 4, __kCFDictionaryHasNullCallBacks);
    } else if (__CFDictionaryValueCallBacksMatchCFType(valueCallBacks)) {
	__CFBitfieldSetValue(flags, 5, 4, __kCFDictionaryHasCFTypeCallBacks);
    } else {
	__CFBitfieldSetValue(flags, 5, 4, __kCFDictionaryHasCustomCallBacks);
    }
    size = __CFDictionaryGetSizeOfType(flags) - sizeof(CFRuntimeBase);
    switch (__CFBitfieldGetValue(flags, 1, 0)) {
    case __kCFDictionaryImmutable:
    case __kCFDictionaryFixedMutable:
	size += __CFDictionaryNumBucketsForCapacity(capacity) * sizeof(struct __CFDictionaryBucket);
	size += __CFDictionaryNumBucketsForCapacity(capacity) * sizeof(const void *);
	break;
    case __kCFDictionaryMutable:
	break;
    }
    memory = (struct __CFDictionary *)_CFRuntimeCreateInstance(allocator, __kCFDictionaryTypeID, size, NULL);
    if (NULL == memory) {
	return NULL;
    }
    __CFBitfieldSetValue(memory->_base._info, 6, 0, flags);
    memory->_count = 0;
    memory->_emptyMarker = (const void *)0xa1b1c1d3;
    memory->_deletedMarker = (const void *)0xa1b1c1d5;
    memory->_context = NULL;
    switch (__CFBitfieldGetValue(flags, 1, 0)) {
    case __kCFDictionaryImmutable:
	if (__CFOASafe) __CFSetLastAllocationEventName(memory, "CFDictionary (immutable)");
	memory->_capacity = capacity;	/* Don't round up capacity */
	memory->_bucketsNum = __CFDictionaryNumBucketsForCapacity(memory->_capacity);
	memory->_buckets = (struct __CFDictionaryBucket *)((uint8_t *)memory + __CFDictionaryGetSizeOfType(flags));
	memory->_values = (const void **)(memory->_buckets + memory->_bucketsNum);
	for (idx = memory->_bucketsNum; idx--;) {
	    memory->_buckets[idx]._key = memory->_emptyMarker;
	}
	break;
    case __kCFDictionaryFixedMutable:
	if (__CFOASafe) __CFSetLastAllocationEventName(memory, "CFDictionary (mutable-fixed)");
	memory->_capacity = capacity;	/* Don't round up capacity */
	memory->_bucketsNum = __CFDictionaryNumBucketsForCapacity(memory->_capacity);
	memory->_buckets = (struct __CFDictionaryBucket *)((uint8_t *)memory + __CFDictionaryGetSizeOfType(flags));
	memory->_values = (const void **)(memory->_buckets + memory->_bucketsNum);
	for (idx = memory->_bucketsNum; idx--;) {
	    memory->_buckets[idx]._key = memory->_emptyMarker;
	}
	break;
    case __kCFDictionaryMutable:
	if (__CFOASafe) __CFSetLastAllocationEventName(memory, "CFDictionary (mutable-variable)");
	memory->_capacity = __CFDictionaryRoundUpCapacity(1);
	memory->_bucketsNum = 0;
	memory->_buckets = NULL;
	memory->_values = NULL;
	break;
    }
    if (__kCFDictionaryHasCustomCallBacks == __CFBitfieldGetValue(flags, 3, 2)) {
	const CFDictionaryKeyCallBacks *cb = __CFDictionaryGetKeyCallBacks((CFDictionaryRef)memory);
	*(CFDictionaryKeyCallBacks *)cb = *callBacks;
	FAULT_CALLBACK((void **)&(cb->retain));
	FAULT_CALLBACK((void **)&(cb->release));
	FAULT_CALLBACK((void **)&(cb->copyDescription));
	FAULT_CALLBACK((void **)&(cb->equal));
	FAULT_CALLBACK((void **)&(cb->hash));
    }
    if (__kCFDictionaryHasCustomCallBacks == __CFBitfieldGetValue(flags, 5, 4)) {
	const CFDictionaryValueCallBacks *vcb = __CFDictionaryGetValueCallBacks((CFDictionaryRef)memory);
	*(CFDictionaryValueCallBacks *)vcb = *valueCallBacks;
	FAULT_CALLBACK((void **)&(vcb->retain));
	FAULT_CALLBACK((void **)&(vcb->release));
	FAULT_CALLBACK((void **)&(vcb->copyDescription));
	FAULT_CALLBACK((void **)&(vcb->equal));
    }
    return (CFDictionaryRef)memory;
}

CFDictionaryRef CFDictionaryCreate(CFAllocatorRef allocator, const void **keys, const void **values, CFIndex numValues, const CFDictionaryKeyCallBacks *keyCallBacks, const CFDictionaryValueCallBacks *valueCallBacks) {
    CFDictionaryRef result;
    UInt32 flags;
    CFIndex idx;
    CFAssert2(0 <= numValues, __kCFLogAssertion, "%s(): numValues (%d) cannot be less than zero", __PRETTY_FUNCTION__, numValues);
    result = __CFDictionaryInit(allocator, __kCFDictionaryImmutable, numValues, keyCallBacks, valueCallBacks);
    flags = __CFBitfieldGetValue(((CFRuntimeBase *)result)->_info, 1, 0);
    if (flags == __kCFDictionaryImmutable) {
        __CFBitfieldSetValue(((CFRuntimeBase *)result)->_info, 1, 0, __kCFDictionaryFixedMutable);
    }
    for (idx = 0; idx < numValues; idx++) {
	CFDictionaryAddValue((CFMutableDictionaryRef)result, keys[idx], values[idx]);
    }
    __CFBitfieldSetValue(((CFRuntimeBase *)result)->_info, 1, 0, flags);
    return result;
}

CFMutableDictionaryRef CFDictionaryCreateMutable(CFAllocatorRef allocator, CFIndex capacity, const CFDictionaryKeyCallBacks *keyCallBacks, const CFDictionaryValueCallBacks *valueCallBacks) {
    CFAssert2(0 <= capacity, __kCFLogAssertion, "%s(): capacity (%d) cannot be less than zero", __PRETTY_FUNCTION__, capacity);
    return (CFMutableDictionaryRef)__CFDictionaryInit(allocator, (0 == capacity) ? __kCFDictionaryMutable : __kCFDictionaryFixedMutable, capacity, keyCallBacks, valueCallBacks);
}

CFDictionaryRef CFDictionaryCreateCopy(CFAllocatorRef allocator, CFDictionaryRef dict) {
    CFDictionaryRef result;
    const CFDictionaryKeyCallBacks *cb;
    const CFDictionaryValueCallBacks *vcb;
    CFIndex numValues = CFDictionaryGetCount(dict);
    const void **list, *buffer[256];
    const void **vlist, *vbuffer[256];
    list = (numValues <= 256) ? buffer : CFAllocatorAllocate(allocator, numValues * sizeof(void *), 0);
    if (list != buffer && __CFOASafe) __CFSetLastAllocationEventName(list, "CFDictionary (temp)");
    vlist = (numValues <= 256) ? vbuffer : CFAllocatorAllocate(allocator, numValues * sizeof(void *), 0);
    if (vlist != vbuffer && __CFOASafe) __CFSetLastAllocationEventName(vlist, "CFDictionary (temp)");
    CFDictionaryGetKeysAndValues(dict, list, vlist);
    cb = CF_IS_OBJC(__kCFDictionaryTypeID, dict) ? &kCFTypeDictionaryKeyCallBacks : __CFDictionaryGetKeyCallBacks(dict);
    vcb = CF_IS_OBJC(__kCFDictionaryTypeID, dict) ? &kCFTypeDictionaryValueCallBacks : __CFDictionaryGetValueCallBacks(dict);
    result = CFDictionaryCreate(allocator, list, vlist, numValues, cb, vcb);
    if (list != buffer) CFAllocatorDeallocate(allocator, list);
    if (vlist != vbuffer) CFAllocatorDeallocate(allocator, vlist);
    return result;
}

CFMutableDictionaryRef CFDictionaryCreateMutableCopy(CFAllocatorRef allocator, CFIndex capacity, CFDictionaryRef dict) {
    CFMutableDictionaryRef result;
    const CFDictionaryKeyCallBacks *cb;
    const CFDictionaryValueCallBacks *vcb;
    CFIndex idx, numValues = CFDictionaryGetCount(dict);
    const void **list, *buffer[256];
    const void **vlist, *vbuffer[256];
    CFAssert3(0 == capacity || numValues <= capacity, __kCFLogAssertion, "%s(): for fixed-mutable dicts, capacity (%d) must be greater than or equal to initial number of values (%d)", __PRETTY_FUNCTION__, capacity, numValues);
    list = (numValues <= 256) ? buffer : CFAllocatorAllocate(allocator, numValues * sizeof(void *), 0);
    if (list != buffer && __CFOASafe) __CFSetLastAllocationEventName(list, "CFDictionary (temp)");
    vlist = (numValues <= 256) ? vbuffer : CFAllocatorAllocate(allocator, numValues * sizeof(void *), 0);
    if (vlist != vbuffer && __CFOASafe) __CFSetLastAllocationEventName(vlist, "CFDictionary (temp)");
    CFDictionaryGetKeysAndValues(dict, list, vlist);
    cb = CF_IS_OBJC(__kCFDictionaryTypeID, dict) ? &kCFTypeDictionaryKeyCallBacks : __CFDictionaryGetKeyCallBacks(dict);
    vcb = CF_IS_OBJC(__kCFDictionaryTypeID, dict) ? &kCFTypeDictionaryValueCallBacks : __CFDictionaryGetValueCallBacks(dict);
    result = CFDictionaryCreateMutable(allocator, capacity, cb, vcb);
    if (0 == capacity) _CFDictionarySetCapacity(result, numValues);
    for (idx = 0; idx < numValues; idx++) {
	CFDictionaryAddValue(result, list[idx], vlist[idx]);
    }
    if (list != buffer) CFAllocatorDeallocate(allocator, list);
    if (vlist != vbuffer) CFAllocatorDeallocate(allocator, vlist);
    return result;
}

void _CFDictionarySetContext(CFDictionaryRef dict, void *context) {
    ((struct __CFDictionary *)dict)->_context = context;
}

CFIndex CFDictionaryGetCount(CFDictionaryRef dict) {
    CF_OBJC_FUNCDISPATCH0(__kCFDictionaryTypeID, CFIndex, dict, "count");
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    return dict->_count;
}

CFIndex CFDictionaryGetCountOfKey(CFDictionaryRef dict, const void *key) {
    CFIndex match;
    CF_OBJC_FUNCDISPATCH1(__kCFDictionaryTypeID, CFIndex, dict, "countForKey:", (id)key);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    if (0 == dict->_count) return 0;
    if (__kCFDictionaryHasNullCallBacks == __CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 3, 2)) {
	__CFDictionaryFindBuckets1a(dict, key, &match);
    } else {
	__CFDictionaryFindBuckets1b(dict, key, &match);
    }
    return (kCFNotFound != match ? 1 : 0);
}

CFIndex CFDictionaryGetCountOfValue(CFDictionaryRef dict, const void *value) {
    struct __CFDictionaryBucket *buckets;
    const CFDictionaryValueCallBacks *vcb;
    CFIndex idx, cnt = 0, nbuckets;
    CF_OBJC_FUNCDISPATCH1(__kCFDictionaryTypeID, CFIndex, dict, "countForObject:", (id)value);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    if (0 == dict->_count) return 0;
    buckets = dict->_buckets;
    nbuckets = dict->_bucketsNum;
    vcb = __CFDictionaryGetValueCallBacks(dict);
    for (idx = 0; idx < nbuckets; idx++) {
	if (__CFDictionaryBucketIsOccupied(dict, &buckets[idx])) {
	    if (dict->_values[idx] != value && vcb->equal && !INVOKE_CALLBACK3(vcb->equal, dict->_values[idx], value, dict->_context)) {
		cnt++;
	    }
	}
    }
    return cnt;
}

Boolean CFDictionaryContainsKey(CFDictionaryRef dict, const void *key) {
    CFIndex match;
    CF_OBJC_FUNCDISPATCH1(__kCFDictionaryTypeID, Boolean, dict, "containsKey:", (id)key);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    if (0 == dict->_count) return false;
    if (__kCFDictionaryHasNullCallBacks == __CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 3, 2)) {
	__CFDictionaryFindBuckets1a(dict, key, &match);
    } else {
	__CFDictionaryFindBuckets1b(dict, key, &match);
    }
    return (kCFNotFound != match ? true : false);
}

Boolean CFDictionaryContainsValue(CFDictionaryRef dict, const void *value) {
    struct __CFDictionaryBucket *buckets;
    const CFDictionaryValueCallBacks *vcb;
    CFIndex idx, nbuckets;
    CF_OBJC_FUNCDISPATCH1(__kCFDictionaryTypeID, Boolean, dict, "containsObject:", (id)value);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    if (0 == dict->_count) return false;
    buckets = dict->_buckets;
    nbuckets = dict->_bucketsNum;
    vcb = __CFDictionaryGetValueCallBacks(dict);
    for (idx = 0; idx < nbuckets; idx++) {
	if (__CFDictionaryBucketIsOccupied(dict, &buckets[idx])) {
	    if (dict->_values[idx] != value && vcb->equal && !INVOKE_CALLBACK3(vcb->equal, dict->_values[idx], value, dict->_context)) {
		return true;
	    }
	}
    }
    return false;
}

const void *CFDictionaryGetValue(CFDictionaryRef dict, const void *key) {
    CFIndex match;
    CF_OBJC_FUNCDISPATCH1(__kCFDictionaryTypeID, const void *, dict, "objectForKey:", key);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    if (0 == dict->_count) return NULL;
    if (__kCFDictionaryHasNullCallBacks == __CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 3, 2)) {
	__CFDictionaryFindBuckets1a(dict, key, &match);
    } else {
	__CFDictionaryFindBuckets1b(dict, key, &match);
    }
    return (kCFNotFound != match ? dict->_values[match] : NULL);
}

Boolean CFDictionaryGetValueIfPresent(CFDictionaryRef dict, const void *key, const void **value) {
    CFIndex match;
    CF_OBJC_FUNCDISPATCH2(__kCFDictionaryTypeID, Boolean, dict, "_getValue:forKey:", (id *)value, key);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    if (0 == dict->_count) return false;
    if (__kCFDictionaryHasNullCallBacks == __CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 3, 2)) {
	__CFDictionaryFindBuckets1a(dict, key, &match);
    } else {
	__CFDictionaryFindBuckets1b(dict, key, &match);
    }
    return (kCFNotFound != match ? ((value ? *value = dict->_values[match] : NULL), true) : false);
}

Boolean CFDictionaryGetKeyIfPresent(CFDictionaryRef dict, const void *key, const void **actualkey) {
    CFIndex match;
//#warning CF: not toll-free bridged
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    if (0 == dict->_count) return false;
    if (__kCFDictionaryHasNullCallBacks == __CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 3, 2)) {
	__CFDictionaryFindBuckets1a(dict, key, &match);
    } else {
	__CFDictionaryFindBuckets1b(dict, key, &match);
    }
    return (kCFNotFound != match ? ((actualkey ? *actualkey = dict->_buckets[match]._key : NULL), true) : false);
}

void CFDictionaryGetKeysAndValues(CFDictionaryRef dict, const void **keys, const void **values) {
    struct __CFDictionaryBucket *buckets;
    CFIndex idx, cnt, nbuckets;
    CF_OBJC_FUNCDISPATCH2(__kCFDictionaryTypeID, void, dict, "getObjects:andKeys:", (id *)values, (id *)keys);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    buckets = dict->_buckets;
    nbuckets = dict->_bucketsNum;
    for (idx = 0; idx < nbuckets; idx++) {
	if (__CFDictionaryBucketIsOccupied(dict, &buckets[idx])) {
	    for (cnt = 1; cnt--;) {
		if (keys) *keys++ = buckets[idx]._key;
		if (values) *values++ = dict->_values[idx];
	    }
	}
    }
}

void CFDictionaryApplyFunction(CFDictionaryRef dict, CFDictionaryApplierFunction applier, void *context) {
    struct __CFDictionaryBucket *buckets;
    CFIndex idx, cnt, nbuckets;
    FAULT_CALLBACK((void **)&(applier));
    CF_OBJC_FUNCDISPATCH2(__kCFDictionaryTypeID, void, dict, "_apply:context:", applier, context);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    buckets = dict->_buckets;
    nbuckets = dict->_bucketsNum;
    for (idx = 0; idx < nbuckets; idx++) {
	if (__CFDictionaryBucketIsOccupied(dict, &buckets[idx])) {
	    for (cnt = 1; cnt--;) {
		INVOKE_CALLBACK3(applier, buckets[idx]._key, dict->_values[idx], context);
	    }
	}
    }
}

static void __CFDictionaryGrow(CFMutableDictionaryRef dict, CFIndex numNewValues) {
    struct __CFDictionaryBucket *oldbuckets = dict->_buckets;
    const void **oldvalues = dict->_values;
    CFIndex idx, oldnbuckets = dict->_bucketsNum;
    CFIndex oldCount = dict->_count;
    dict->_capacity = __CFDictionaryRoundUpCapacity(oldCount + numNewValues);
    dict->_bucketsNum = __CFDictionaryNumBucketsForCapacity(dict->_capacity);
    dict->_buckets = CFAllocatorAllocate(CFGetAllocator(dict), dict->_bucketsNum * sizeof(struct __CFDictionaryBucket) + dict->_bucketsNum * sizeof(const void *), 0);
    dict->_values = (const void **)(dict->_buckets + dict->_bucketsNum);
    if (NULL == dict->_buckets) __CFAbort();
    if (__CFOASafe) __CFSetLastAllocationEventName(dict->_buckets, "CFDictionary (store)");
    for (idx = dict->_bucketsNum; idx--;) {
	dict->_buckets[idx]._key = dict->_emptyMarker;
    }
    if (NULL == oldbuckets) return;
    for (idx = 0; idx < oldnbuckets; idx++) {
	if (__CFDictionaryBucketIsOccupied(dict, &oldbuckets[idx])) {
	    CFIndex match, nomatch;
	    __CFDictionaryFindBuckets2(dict, oldbuckets[idx]._key, &match, &nomatch);
	    CFAssert3(kCFNotFound == match, __kCFLogAssertion, "%s(): two values (%p, %p) now hash to the same slot; mutable value changed while in table or hash value is not immutable", __PRETTY_FUNCTION__, oldbuckets[idx]._key, dict->_buckets[match]._key);
	    dict->_buckets[nomatch]._key = oldbuckets[idx]._key;
	    dict->_values[nomatch] = oldvalues[idx];
	}
    }
    CFAssert1(dict->_count == oldCount, __kCFLogAssertion, "%s(): dict count differs after rehashing; error", __PRETTY_FUNCTION__);
    CFAllocatorDeallocate(CFGetAllocator(dict), oldbuckets);
}

// This function is for Foundation's benefit; no one else should use it.
void _CFDictionarySetCapacity(CFMutableDictionaryRef dict, CFIndex cap) {
    if (CF_IS_OBJC(__kCFDictionaryTypeID, dict)) return;
#if defined(DEBUG)
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    CFAssert1(__CFDictionaryGetType(dict) != __kCFDictionaryImmutable && __CFDictionaryGetType(dict) != __kCFDictionaryFixedMutable, __kCFLogAssertion, "%s(): dict is immutable or fixed-mutable", __PRETTY_FUNCTION__);
    CFAssert3(dict->_count <= cap, __kCFLogAssertion, "%s(): desired capacity (%d) is less than count (%d)", __PRETTY_FUNCTION__, cap, dict->_count);
#endif
    __CFDictionaryGrow(dict, cap - dict->_count);
}

// This function is for Foundation's benefit; no one else should use it.
Boolean _CFDictionaryIsMutable(CFDictionaryRef dict) {
    return (__CFDictionaryGetType(dict) != __kCFDictionaryImmutable);
}   

void CFDictionaryAddValue(CFMutableDictionaryRef dict, const void *key, const void *value) {
    CFIndex match, nomatch;
    const CFDictionaryKeyCallBacks *cb;
    const CFDictionaryValueCallBacks *vcb;
    const void *newKey, *newValue;
    CF_OBJC_FUNCDISPATCH2(__kCFDictionaryTypeID, void, dict, "_addObject:forKey:", value, key);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    switch (__CFDictionaryGetType(dict)) {
    case __kCFDictionaryMutable:
	if (dict->_count == dict->_capacity || NULL == dict->_buckets) {
	    __CFDictionaryGrow(dict, 1);
	}
	break;
    case __kCFDictionaryFixedMutable:
	CFAssert3(dict->_count < dict->_capacity, __kCFLogAssertion, "%s(): capacity exceeded on fixed-capacity dict %p (capacity = %d)", __PRETTY_FUNCTION__, dict, dict->_capacity);
	break;
    default:
	CFAssert2(__CFDictionaryGetType(dict) != __kCFDictionaryImmutable, __kCFLogAssertion, "%s(): immutable dict %p passed to mutating operation", __PRETTY_FUNCTION__, dict);
	break;
    }
    __CFDictionaryFindBuckets2(dict, key, &match, &nomatch);
    if (kCFNotFound != match) {
    } else {
	cb = __CFDictionaryGetKeyCallBacks(dict);
	vcb = __CFDictionaryGetValueCallBacks(dict);
	if (cb->retain) {
	    newKey = (void *)INVOKE_CALLBACK3(cb->retain, CFGetAllocator(dict), key, dict->_context);
	} else {
	    newKey = key;
	}
	if (vcb->retain) {
	    newValue = (void *)INVOKE_CALLBACK3(vcb->retain, CFGetAllocator(dict), value, dict->_context);
	} else {
	    newValue = value;
	}
	if (dict->_emptyMarker == newKey) {
	    __CFDictionaryFindNewEmptyMarker(dict);
	}
	if (dict->_deletedMarker == newKey) {
	    __CFDictionaryFindNewDeletedMarker(dict);
	}
	dict->_buckets[nomatch]._key = newKey;
	dict->_values[nomatch] = newValue;
	dict->_count++;
    }
}

void CFDictionaryReplaceValue(CFMutableDictionaryRef dict, const void *key, const void *value) {
    CFIndex match;
    const CFDictionaryValueCallBacks *vcb;
    const void *newValue;
    CF_OBJC_FUNCDISPATCH2(__kCFDictionaryTypeID, void, dict, "_replaceObject:forKey:", value, key);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    switch (__CFDictionaryGetType(dict)) {
    case __kCFDictionaryMutable:
    case __kCFDictionaryFixedMutable:
	break;
    default:
	CFAssert2(__CFDictionaryGetType(dict) != __kCFDictionaryImmutable, __kCFLogAssertion, "%s(): immutable dict %p passed to mutating operation", __PRETTY_FUNCTION__, dict);
	break;
    }
    if (0 == dict->_count) return;
    if (__kCFDictionaryHasNullCallBacks == __CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 3, 2)) {
	__CFDictionaryFindBuckets1a(dict, key, &match);
    } else {
	__CFDictionaryFindBuckets1b(dict, key, &match);
    }
    if (kCFNotFound == match) return;
    vcb = __CFDictionaryGetValueCallBacks(dict);
    if (vcb->retain) {
	newValue = (void *)INVOKE_CALLBACK3(vcb->retain, CFGetAllocator(dict), value, dict->_context);
    } else {
	newValue = value;
    }
    if (vcb->release) {
	INVOKE_CALLBACK3(vcb->release, CFGetAllocator(dict), dict->_values[match], dict->_context);
    }
    dict->_values[match] = newValue;
}

void CFDictionarySetValue(CFMutableDictionaryRef dict, const void *key, const void *value) {
    CFIndex match, nomatch;
    const CFDictionaryKeyCallBacks *cb;
    const CFDictionaryValueCallBacks *vcb;
    const void *newKey, *newValue;
    CF_OBJC_FUNCDISPATCH2(__kCFDictionaryTypeID, void, dict, "setObject:forKey:", value, key);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    switch (__CFDictionaryGetType(dict)) {
    case __kCFDictionaryMutable:
	if (dict->_count == dict->_capacity || NULL == dict->_buckets) {
	    __CFDictionaryGrow(dict, 1);
	}
	break;
    case __kCFDictionaryFixedMutable:
	break;
    default:
	CFAssert2(__CFDictionaryGetType(dict) != __kCFDictionaryImmutable, __kCFLogAssertion, "%s(): immutable dict %p passed to mutating operation", __PRETTY_FUNCTION__, dict);
	break;
    }
    __CFDictionaryFindBuckets2(dict, key, &match, &nomatch);
    vcb = __CFDictionaryGetValueCallBacks(dict);
    if (vcb->retain) {
	newValue = (void *)INVOKE_CALLBACK3(vcb->retain, CFGetAllocator(dict), value, dict->_context);
    } else {
	newValue = value;
    }
    if (kCFNotFound != match) {
	if (vcb->release) {
	    INVOKE_CALLBACK3(vcb->release, CFGetAllocator(dict), dict->_values[match], dict->_context);
	}
	dict->_values[match] = newValue;
    } else {
	CFAssert3(__kCFDictionaryFixedMutable != __CFDictionaryGetType(dict) || dict->_count < dict->_capacity, __kCFLogAssertion, "%s(): capacity exceeded on fixed-capacity dict %p (capacity = %d)", __PRETTY_FUNCTION__, dict, dict->_capacity);
	cb = __CFDictionaryGetKeyCallBacks(dict);
	if (cb->retain) {
	    newKey = (void *)INVOKE_CALLBACK3(cb->retain, CFGetAllocator(dict), key, dict->_context);
	} else {
	    newKey = key;
	}
	if (dict->_emptyMarker == newKey) {
	    __CFDictionaryFindNewEmptyMarker(dict);
	}
	if (dict->_deletedMarker == newKey) {
	    __CFDictionaryFindNewDeletedMarker(dict);
	}
	dict->_buckets[nomatch]._key = newKey;
	dict->_values[nomatch] = newValue;
	dict->_count++;
    }
}

void CFDictionaryRemoveValue(CFMutableDictionaryRef dict, const void *key) {
    CFIndex match;
    const CFDictionaryKeyCallBacks *cb;
    const CFDictionaryValueCallBacks *vcb;
    CF_OBJC_FUNCDISPATCH1(__kCFDictionaryTypeID, void, dict, "removeObjectForKey:", key);
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    switch (__CFDictionaryGetType(dict)) {
    case __kCFDictionaryMutable:
    case __kCFDictionaryFixedMutable:
	break;
    default:
	CFAssert2(__CFDictionaryGetType(dict) != __kCFDictionaryImmutable, __kCFLogAssertion, "%s(): immutable dict %p passed to mutating operation", __PRETTY_FUNCTION__, dict);
	break;
    }
    if (0 == dict->_count) return;
    if (__kCFDictionaryHasNullCallBacks == __CFBitfieldGetValue(((CFRuntimeBase *)dict)->_info, 3, 2)) {
	__CFDictionaryFindBuckets1a(dict, key, &match);
    } else {
	__CFDictionaryFindBuckets1b(dict, key, &match);
    }
    if (kCFNotFound == match) return;
    dict->_count--;
    if (1) {
	cb = __CFDictionaryGetKeyCallBacks(dict);
	vcb = __CFDictionaryGetValueCallBacks(dict);
	if (cb->release) {
	    INVOKE_CALLBACK3(cb->release, CFGetAllocator(dict), dict->_buckets[match]._key, dict->_context);
	}
	if (vcb->release) {
	    INVOKE_CALLBACK3(vcb->release, CFGetAllocator(dict), dict->_values[match], dict->_context);
	}
        dict->_buckets[match]._key = dict->_deletedMarker;
    }
}

void CFDictionaryRemoveAllValues(CFMutableDictionaryRef dict) {
    struct __CFDictionaryBucket *buckets;
    const CFDictionaryKeyCallBacks *cb;
    const CFDictionaryValueCallBacks *vcb;
    CFAllocatorRef allocator;
    CFIndex idx, nbuckets;
    CF_OBJC_FUNCDISPATCH0(__kCFDictionaryTypeID, void, dict, "removeAllObjects");
    __CFGenericValidateType(dict, __kCFDictionaryTypeID);
    switch (__CFDictionaryGetType(dict)) {
    case __kCFDictionaryMutable:
    case __kCFDictionaryFixedMutable:
	break;
    default:
	CFAssert2(__CFDictionaryGetType(dict) != __kCFDictionaryImmutable, __kCFLogAssertion, "%s(): immutable dict %p passed to mutating operation", __PRETTY_FUNCTION__, dict);
	break;
    }
    if (0 == dict->_count) return;
    buckets = dict->_buckets;
    nbuckets = dict->_bucketsNum;
    cb = __CFDictionaryGetKeyCallBacks(dict);
    vcb = __CFDictionaryGetValueCallBacks(dict);
    allocator = CFGetAllocator(dict);
    for (idx = 0; idx < nbuckets; idx++) {
	if (__CFDictionaryBucketIsOccupied(dict, &buckets[idx])) {
	    if (cb->release) {
		INVOKE_CALLBACK3(cb->release, allocator, buckets[idx]._key, dict->_context);
	    }
	    if (vcb->release) {
		INVOKE_CALLBACK3(vcb->release, allocator, dict->_values[idx], dict->_context);
	    }
	    buckets[idx]._key = dict->_emptyMarker;
	}
    }
    dict->_count = 0;
}

