/*
 * Copyright (c) 2001 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*	CFDate.c
	Copyright 1998-2001, Apple, Inc. All rights reserved.
	Responsibility: Christopher Kane
*/

#include <CoreFoundation/CFDate.h>
#include <CoreFoundation/CFTimeZone.h>
#include <CoreFoundation/CFDictionary.h>
#include <CoreFoundation/CFArray.h>
#include <CoreFoundation/CFString.h>
#include "CFVeryPrivate.h"
#include "CFInternal.h"
#include <math.h>
#if defined(__MACH__) || defined(__LINUX__)
    #include <sys/time.h>
#endif
#if defined(__MACOS8__)
    #include <DateTimeUtils.h>
#endif
#if defined(__WIN32__)
    #include <windows.h>
#endif

const CFTimeInterval kCFAbsoluteTimeIntervalSince1970 = 978307200.0L;
const CFTimeInterval kCFAbsoluteTimeIntervalSince1904 = 3061152000.0L;

/* cjk: The Julian Date for the reference date is 2451910.5,
        I think, in case that's ever useful. */

struct __CFDate {
    CFRuntimeBase _base;
    CFAbsoluteTime _time;	/* immutable */
};

#if defined(__MACH__)
static double __CFTSRRate = 0.0;
static CFAbsoluteTime __CFBootAbsoluteTime = 0.0;
static CFTimeInterval __CFLastSyncOffset = -1.0E+20;

__private_extern__ int64_t __CFTimeIntervalToTSR(CFTimeInterval ti) {
    return (int64_t)(ti * __CFTSRRate);
}

__private_extern__ CFTimeInterval __CFTSRToTimeInterval(int64_t tsr) {
    return (CFTimeInterval)((double)tsr / __CFTSRRate);
}

__private_extern__ int64_t __CFAbsoluteTimeToTSR(CFAbsoluteTime at) {
    CFTimeInterval delta = __CFReadTSR() / __CFTSRRate;
    if (__CFLastSyncOffset + 0.5 < delta) {
	struct timeval tv;
	/* 0.5 seconds is arbitrarily chosen, but keeps the error under
	0.0003 seconds generally; we need a number which is large enough
	to cut down on the number of gettimeofday() calls, but small
	enough that radical changes to the calendar clock are noticed
	reasonably quickly. */
	gettimeofday(&tv, NULL);
	__CFLastSyncOffset = delta;
	__CFBootAbsoluteTime = ((double)tv.tv_sec - kCFAbsoluteTimeIntervalSince1970) + 1.0E-6 * (double)tv.tv_usec - delta;
    }
    return (at - __CFBootAbsoluteTime ) * __CFTSRRate;
}

__private_extern__ CFAbsoluteTime __CFTSRToAbsoluteTime(int64_t tsr) {
    CFTimeInterval delta = __CFReadTSR() / __CFTSRRate;
    if (__CFLastSyncOffset + 0.5 < delta) {
	struct timeval tv;
	/* 0.5 seconds is arbitrarily chosen, but keeps the error under
	0.0003 seconds generally; we need a number which is large enough
	to cut down on the number of gettimeofday() calls, but small
	enough that radical changes to the calendar clock are noticed
	reasonably quickly. */
	gettimeofday(&tv, NULL);
	__CFLastSyncOffset = delta;
	__CFBootAbsoluteTime = ((double)tv.tv_sec - kCFAbsoluteTimeIntervalSince1970) + 1.0E-6 * (double)tv.tv_usec - delta;
    }
    return __CFBootAbsoluteTime + tsr / __CFTSRRate;
}
#endif

CFAbsoluteTime CFAbsoluteTimeGetCurrent(void) {
    CFAbsoluteTime ret;
#if defined(__WIN32__)
    FILETIME ft;
    GetSystemTimeAsFileTime(&ft);
    ret = (CFTimeInterval)ft.dwHighDateTime * 429.49672960;
    ret += (CFTimeInterval)ft.dwLowDateTime / 10000000.0;
    ret -= (11644473600.0 + kCFAbsoluteTimeIntervalSince1970);
	 /* seconds between 1601 and 1970, 1970 and 2001 */
#endif
#if defined(__MACOS8__)
    /* Resolution only to 1 second */
    unsigned long time;
    GetDateTime(&time);
    ret = (CFTimeInterval)time - kCFAbsoluteTimeIntervalSince1904;
#endif
#if defined(__MACH__)
    ret = __CFTSRToAbsoluteTime(__CFReadTSR());
#endif __MACH__
#if defined(__svr4__) || defined(__hpux__) || defined(__LINUX__)
    struct timeval tv;
    gettimeofday(&tv, NULL);
    ret = (CFTimeInterval)tv.tv_sec - kCFAbsoluteTimeIntervalSince1970;
    ret += (1.0E-6 * (CFTimeInterval)tv.tv_usec);
#endif
    return ret;
}

static Boolean __CFDateEqual(CFTypeRef cf1, CFTypeRef cf2) {
    CFDateRef date1 = (CFDateRef)cf1;
    CFDateRef date2 = (CFDateRef)cf2;
    if (date1->_time != date2->_time) return false;
    return true;
}

static CFHashCode __CFDateHash(CFTypeRef cf) {
    CFDateRef date = (CFDateRef)cf;
    return (CFHashCode)floor(date->_time);
}

static CFStringRef __CFDateCopyDescription(CFTypeRef cf) {
    CFDateRef date = (CFDateRef)cf;
    return CFStringCreateWithFormat(CFGetAllocator(date), NULL, CFSTR("<CFDate 0x%x [0x%x]>{time = %0.09g}"), (UInt32)cf, (UInt32)CFGetAllocator(date), date->_time);
}

static CFTypeID __kCFDateTypeID = _kCFRuntimeNotATypeID;

static const CFRuntimeClass __CFDateClass = {
    0,
    "CFDate",
    NULL,	// init
    NULL,	// copy
    NULL,	// dealloc
    __CFDateEqual,
    __CFDateHash,
    NULL,	// 
    __CFDateCopyDescription
};

__private_extern__ void __CFDateInitialize(void) {
#if defined(__MACH__)
    struct timeval tv;
    struct mach_timebase_info info;
    mach_timebase_info(&info);
    __CFTSRRate = (1.0E9 / (double)info.numer) * (double)info.denom;
#endif
    __kCFDateTypeID = _CFRuntimeRegisterClass(&__CFDateClass);
}

CFTypeID CFDateGetTypeID(void) {
    return __kCFDateTypeID;
}

CFDateRef CFDateCreate(CFAllocatorRef allocator, CFAbsoluteTime at) {
    CFDateRef memory;
    UInt32 size;
    size = sizeof(struct __CFDate) - sizeof(CFRuntimeBase);
    memory = _CFRuntimeCreateInstance(allocator, __kCFDateTypeID, size, NULL);
    if (NULL == memory) {
        return NULL;
    }
    if (__CFOASafe) __CFSetLastAllocationEventName((void *)memory, "CFDate");
    ((struct __CFDate *)memory)->_time = at;
    return memory;
}

CFTimeInterval CFDateGetAbsoluteTime(CFDateRef date) {
    CF_OBJC_FUNCDISPATCH0(__kCFDateTypeID, CFTimeInterval, date, "timeIntervalSinceReferenceDate");
    __CFGenericValidateType(date, CFDateGetTypeID());
    return date->_time;
}

CFTimeInterval CFDateGetTimeIntervalSinceDate(CFDateRef date, CFDateRef otherDate) {
    CF_OBJC_FUNCDISPATCH1(__kCFDateTypeID, CFTimeInterval, date, "timeIntervalSinceDate:", otherDate);
    __CFGenericValidateType(date, CFDateGetTypeID());
    __CFGenericValidateType(otherDate, CFDateGetTypeID());
    return date->_time - otherDate->_time;
}

CFComparisonResult CFDateCompare(CFDateRef date, CFDateRef otherDate, void *context) {
    CF_OBJC_FUNCDISPATCH1(__kCFDateTypeID, CFComparisonResult, date, "compare:", otherDate);
    __CFGenericValidateType(date, CFDateGetTypeID());
    __CFGenericValidateType(otherDate, CFDateGetTypeID());
    if (date->_time < otherDate->_time) return kCFCompareLessThan;
    if (date->_time > otherDate->_time) return kCFCompareGreaterThan;
    return kCFCompareEqualTo;
}

CF_INLINE SInt32 __CFDoubleModToInt(double d, SInt32 modulus) {
    SInt32 result = (SInt32)floor(d - floor(d / modulus) * modulus);
    if (result < 0) result += modulus;
    return result;
}

CF_INLINE double __CFDoubleMod(double d, SInt32 modulus) {
    double result = d - floor(d / modulus) * modulus;
    if (result < 0.0) result += (double)modulus;
    return result;
}

/* year arg is absolute year; Gregorian 2001 == year 0; 2001/1/1 = absolute date 0 */
static SInt32 __CFDaysInMonth(SInt32 month, SInt32 year) {
    switch (month) {
    case 1: return 31;
    case 2: {
	SInt32 y;
	year += 1;	/* correct to nearest multiple-of-400 year */
	y = year % 400;
	return (year % 4 == 0 && y != 100 && y != 200 && y != 300) ? 29 : 28;
    }
    case 3: return 31;
    case 4: return 30;
    case 5: return 31;
    case 6: return 30;
    case 7: return 31;
    case 8: return 31;
    case 9: return 30;
    case 10: return 31;
    case 11: return 30;
    case 12: return 31;
    }
    return 0;
}

/* year arg is absolute year; Gregorian 2001 == year 0; 2001/1/1 = absolute date 0 */
static SInt32 __CFDaysBeforeMonth(SInt32 month, SInt32 year) {
    switch (month) {
    case 1: return 0;
    case 2: return 31;
    case 3: return 31 + __CFDaysInMonth(2, year);
    case 4: return 62 + __CFDaysInMonth(2, year);
    case 5: return 92 + __CFDaysInMonth(2, year);
    case 6: return 123 + __CFDaysInMonth(2, year);
    case 7: return 153 + __CFDaysInMonth(2, year);
    case 8: return 184 + __CFDaysInMonth(2, year);
    case 9: return 215 + __CFDaysInMonth(2, year);
    case 10: return 245 + __CFDaysInMonth(2, year);
    case 11: return 276 + __CFDaysInMonth(2, year);
    case 12: return 306 + __CFDaysInMonth(2, year);
    case 13: return 337 + __CFDaysInMonth(2, year);
    }
    return 0;
}

/* year arg is absolute year; Gregorian 2001 == year 0; 2001/1/1 = absolute date 0 */
static SInt32 __CFDaysAfterMonth(SInt32 month, SInt32 year) {
    switch (month) {
    case 0: return 337 + __CFDaysInMonth(2, year);
    case 1: return 306 + __CFDaysInMonth(2, year);
    case 2: return 306;
    case 3: return 275;
    case 4: return 245;
    case 5: return 214;
    case 6: return 184;
    case 7: return 153;
    case 8: return 122;
    case 9: return 92;
    case 10: return 61;
    case 11: return 31;
    case 12: return 0;
    }
    return 0;
}

/* year arg is absolute year; Gregorian 2001 == year 0; 2001/1/1 = absolute date 0 */
static void __CFYMDFromAbsolute(SInt32 absolute, SInt32 *year, SInt32 *month, SInt32 *day) {
    SInt32 y = 0, ydays;
    while (absolute < 0) {
	y -= 1;
	absolute += __CFDaysAfterMonth(0, y);
    }
    /* Now absolute is non-negative days to add to year */
    ydays = __CFDaysAfterMonth(0, y);
    while (ydays <= absolute) {
	y += 1;
	absolute -= ydays;
	ydays = __CFDaysAfterMonth(0, y);
    }
    /* Now we have year and days-into-year */
    if (year) *year = y;
    if (month || day) {
	SInt32 m = absolute / 33 + 1; /* search from the approximation */
	while (__CFDaysBeforeMonth(m + 1, y) <= absolute) m++;
	if (month) *month = m;
	if (day) *day = absolute - __CFDaysBeforeMonth(m, y) + 1;
    }
}

/* year arg is absolute year; Gregorian 2001 == year 0; 2001/1/1 = absolute date 0 */
static SInt32 __CFAbsoluteFromYMD(SInt32 year, SInt32 month, SInt32 day) {
    SInt32 absolute = 0, idx;
    if (year < 0) {
	for (idx = year; idx < 0; idx++)
	    absolute -= __CFDaysAfterMonth(0, idx);
    } else {
	for (idx = year - 1; 0 <= idx; idx--)
	    absolute += __CFDaysAfterMonth(0, idx);
    }
    /* Now add the days into the original year */
    absolute += __CFDaysBeforeMonth(month, year) + day - 1;
    return absolute;
}

Boolean CFGregorianDateIsValid(CFGregorianDate gdate, CFOptionFlags unitFlags) {
    if (gdate.year <= 0) return false;
    if (gdate.month < 1 || 12 < gdate.month) return false;
    if (gdate.day < 1 || 31 < gdate.day) return false;
    if (gdate.hour < 0 || 24 <= gdate.hour) return false;
    if (gdate.minute < 0 || 60 <= gdate.minute) return false;
    if (gdate.second < 0.0 || 60.0 <= gdate.second) return false;
    if (__CFDaysInMonth(gdate.month, gdate.year - 2001) < gdate.day) return false;
    return true;
}

__private_extern__ Boolean CFGregorianDateValidate(CFGregorianDate gdate, CFOptionFlags unitFlags) {
// Remove this obsolete function someday
    return CFGregorianDateIsValid(gdate, unitFlags);
}

CFAbsoluteTime CFGregorianDateGetAbsoluteTime(CFGregorianDate gdate, CFTimeZoneRef tz) {
    CFAbsoluteTime at;
    CFTimeInterval offset0, offset1;
    if (NULL != tz) {
	__CFGenericValidateType(tz, CFTimeZoneGetTypeID());
    }
    at = 86400.0 * __CFAbsoluteFromYMD(gdate.year - 2001, gdate.month, gdate.day);
    at += 3600.0 * gdate.hour + 60.0 * gdate.minute + gdate.second;
    if (NULL != tz) {
	offset0 = CFTimeZoneGetSecondsFromGMT(tz, at);
	offset1 = CFTimeZoneGetSecondsFromGMT(tz, at - offset0);
	at -= (offset0 != offset1) ? offset1 : offset0;
    }
    return at;
}

CFGregorianDate CFAbsoluteTimeGetGregorianDate(CFAbsoluteTime at, CFTimeZoneRef tz) {
    CFGregorianDate gdate;
    SInt32 absolute, year, month, day;
    CFAbsoluteTime fixedat;
    if (NULL != tz) {
	__CFGenericValidateType(tz, CFTimeZoneGetTypeID());
    }
    fixedat = at + (NULL != tz ? CFTimeZoneGetSecondsFromGMT(tz, at) : 0.0);
    absolute = (SInt32)floor(fixedat / 86400.0);
    __CFYMDFromAbsolute(absolute, &year, &month, &day);
    gdate.year = year + 2001;
    gdate.month = month;
    gdate.day = day;
    gdate.hour = __CFDoubleModToInt(floor(fixedat / 3600.0), 24);
    gdate.minute = __CFDoubleModToInt(floor(fixedat / 60.0), 60);
    gdate.second = __CFDoubleMod(fixedat, 60);
    return gdate;
}

/* Note that the units of years and months are not equal length, but are treated as such. */
CFAbsoluteTime CFAbsoluteTimeAddGregorianUnits(CFAbsoluteTime at, CFTimeZoneRef tz, CFGregorianUnits units) {
    CFGregorianDate gdate;
    CFGregorianUnits working;
    CFAbsoluteTime candidate_at0, candidate_at1;
    SInt32 monthdays;

    if (NULL != tz) {
	__CFGenericValidateType(tz, CFTimeZoneGetTypeID());
    }

    /* Most people seem to expect years, then months, then days, etc.
	to be added in that order.  Thus, 27 April + (4 days, 1 month)
	= 31 May, and not 1 June. This is also relatively predictable.

	On another issue, months not being equal length, people also
	seem to expect late day-of-month clamping (don't clamp as you
	go through months), but clamp before adding in the days. Late
	clamping is also more predictable given random starting points
	and random numbers of months added (ie Jan 31 + 2 months could
	be March 28 or March 29 in different years with aggressive
	clamping). Proportionality (28 Feb + 1 month = 31 March) is
	also not expected.

	Also, people don't expect time zone transitions to have any
	effect when adding years and/or months and/or days, only.
	Hours, minutes, and seconds, though, are added in as humans
	woulds experience the passing of that time. What this means
	is that if the date, after adding years, months, and days
	lands on some date, and then adding hours, minutes, and
	seconds crosses a time zone transition, the time zone
	transition is accounted for. If adding years, months, and
	days gets the date into a different time zone offset period,
	that transition is not taken into account.
    */
    gdate = CFAbsoluteTimeGetGregorianDate(at, tz);
    /* We must work in a CFGregorianUnits, because the fields in the CFGregorianDate can easily overflow */
    working.years = gdate.year;
    working.months = gdate.month;
    working.days = gdate.day;
    working.years += units.years;
    working.months += units.months;
    while (12 < working.months) {
	working.months -= 12;
	working.years += 1;
    }
    while (working.months < 1) {
	working.months += 12;
	working.years -= 1;
    }
    monthdays = __CFDaysInMonth(working.months, working.years - 2001);
    if (monthdays < working.days) {	/* Clamp day to new month */
	working.days = monthdays;
    }
    working.days += units.days;
    while (monthdays < working.days) {
	working.months += 1;
	if (12 < working.months) {
	    working.months -= 12;
	    working.years += 1;
	}
	working.days -= monthdays;
	monthdays = __CFDaysInMonth(working.months, working.years - 2001);
    }
    while (working.days < 1) {
	working.months -= 1;
	if (working.months < 1) {
	    working.months += 12;
	    working.years -= 1;
	}
	monthdays = __CFDaysInMonth(working.months, working.years - 2001);
	working.days += monthdays;
    }
    gdate.year = working.years;
    gdate.month = working.months;
    gdate.day = working.days;
    /* Roll in hours, minutes, and seconds */
    candidate_at0 = CFGregorianDateGetAbsoluteTime(gdate, tz);
    candidate_at1 = candidate_at0 + 3600.0 * units.hours + 60.0 * units.minutes + units.seconds;
    /* If summing in the hours, minutes, and seconds delta pushes us
     * into a new time zone offset, that will automatically be taken
     * care of by the fact that we just add the raw time above. To
     * undo that effect, we'd have to get the time zone offsets for
     * candidate_at0 and candidate_at1 here, and subtract the
     * difference (offset1 - offset0) from candidate_at1. */
    return candidate_at1;
}

/* at1 - at2.  The only constraint here is that this needs to be the inverse
of CFAbsoluteTimeByAddingGregorianUnits(), but that's a very rigid constraint.
Unfortunately, due to the nonuniformity of the year and month units, this
inversion essentially has to approximate until it finds the answer. */
CFGregorianUnits CFAbsoluteTimeGetDifferenceAsGregorianUnits(CFAbsoluteTime at1, CFAbsoluteTime at2, CFTimeZoneRef tz, CFOptionFlags unitFlags) {
    const SInt32 seconds[5] = {366 * 24 * 3600, 31 * 24 * 3600, 24 * 3600, 3600, 60};
    CFGregorianUnits units = {0, 0, 0, 0, 0, 0.0};
    CFAbsoluteTime atold, atnew = at2;
    SInt32 idx, incr;
    if (at1 == at2) {
	return units;
    }
    incr = (at2 < at1) ? 1 : -1;
    /* Successive approximation: years, then months, then days, then hours, then minutes. */
    for (idx = 0; idx < 5; idx++) {
	if (unitFlags & (1 << idx)) {
	    ((SInt32 *)&units)[idx] = -3 * incr + (at1 - atnew) / seconds[idx];
	    do {
		atold = atnew;
		((SInt32 *)&units)[idx] += incr;
		atnew = CFAbsoluteTimeAddGregorianUnits(at2, tz, units);
	    } while ((1 == incr && atnew <= at1) || (-1 == incr && at1 <= atnew));
	    ((SInt32 *)&units)[idx] -= incr;
	    atnew = atold;
	}
    }
    if (unitFlags & kCFGregorianUnitsSeconds) {
	units.seconds = at1 - atnew;
    }
    return units;
}

SInt32 CFAbsoluteTimeGetDayOfWeek(CFAbsoluteTime at, CFTimeZoneRef tz) {
    SInt32 absolute;
    CFAbsoluteTime fixedat;
    if (NULL != tz) {
	__CFGenericValidateType(tz, CFTimeZoneGetTypeID());
    }
    fixedat = at + (NULL != tz ? CFTimeZoneGetSecondsFromGMT(tz, at) : 0.0);
    absolute = (SInt32)floor(fixedat / 86400.0);
    return (absolute < 0) ? ((absolute + 1) % 7 + 7) : (absolute % 7 + 1); /* Monday = 1, etc. */
}

SInt32 CFAbsoluteTimeGetDayOfYear(CFAbsoluteTime at, CFTimeZoneRef tz) {
    SInt32 absolute, year, month, day;
    CFAbsoluteTime fixedat;
    if (NULL != tz) {
	__CFGenericValidateType(tz, CFTimeZoneGetTypeID());
    }
    fixedat = at + (NULL != tz ? CFTimeZoneGetSecondsFromGMT(tz, at) : 0.0);
    absolute = (SInt32)floor(fixedat / 86400.0);
    __CFYMDFromAbsolute(absolute, &year, &month, &day);
    return __CFDaysBeforeMonth(month, year) + day;
}

/* "the first week of a year is the one which includes the first Thursday" (ISO 8601) */
SInt32 CFAbsoluteTimeGetWeekOfYear(CFAbsoluteTime at, CFTimeZoneRef tz) {
    SInt32 absolute, year, month, day, absolute0101, dow0101;
    CFAbsoluteTime fixedat;
    if (NULL != tz) {
	__CFGenericValidateType(tz, CFTimeZoneGetTypeID());
    }
    fixedat = at + (NULL != tz ? CFTimeZoneGetSecondsFromGMT(tz, at) : 0.0);
    absolute = (SInt32)floor(fixedat / 86400.0);
    __CFYMDFromAbsolute(absolute, &year, &month, &day);
    absolute0101 = __CFAbsoluteFromYMD(year, 1, 1);
    dow0101 = (absolute0101 < 0) ? ((absolute0101 + 1) % 7 + 7) : (absolute0101 % 7 + 1);
    /* First three and last three days of a year can end up in a week of a different year */
    if (1 == month && day < 4) {
	if ((day < 4 && 5 == dow0101) || (day < 3 && 6 == dow0101) || (day < 2 && 7 == dow0101)) {
	    return 53;
	}
    }
    if (12 == month && 28 < day) {
	SInt32 absolute20101, dow20101;
	absolute20101 = __CFAbsoluteFromYMD(year + 1, 1, 1);
	dow20101 = (absolute20101 < 0) ? ((absolute20101 + 1) % 7 + 7) : (absolute20101 % 7 + 1);
	if ((28 < day && 4 == dow20101) || (29 < day && 3 == dow20101) || (30 < day && 2 == dow20101)) {
	    return 1;
	}
    }
    /* Days into year, plus a week-shifting correction, divided by 7. First week is #1. */
    return (__CFDaysBeforeMonth(month, year) + day + (dow0101 - 11) % 7 + 2) / 7 + 1;
}


#if 0

const CFStringRef kCFMonthAbbreviationArray = NULL;
const CFStringRef kCFMonthNameArray = NULL;
const CFStringRef kCFAMPMDesignation = NULL;
const CFStringRef kCFShortDayNameArray = NULL;
const CFStringRef kCFDayNameArray = NULL;


CFStringRef CFGregorianDateCreateStringWithFormat(CFAllocatorRef allocator, CFDictionaryRef formatOptions, CFStringRef format, CFGregorianDate gdate, CFTimeZoneRef tz, ...) {
    static const char *edown[] = {"Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"};
    static const char *edowa[] = {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"};
    static const char *emn[] = {"January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"};
    static const char *ema[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
    static const char *eapa[] = {"AM", "PM"};
    CFMutableStringRef result;
    CFIndex length, curPos;
    CFAbsoluteTime at;
    va_list vargs;

    if (!CFGregorianDateIsValid(gdate, kCFGregorianAllUnits)) {
        return NULL;
    }
    va_start(vargs, tz);
    at = CFGregorianDateGetAbsoluteTime(gdate, tz);
    result = CFStringCreateMutable(allocator, 0);
    length = CFStringGetLength(format);
    for (curPos = 0; curPos < length; curPos++) {
	bool zeroFlag, spaceFlag, prefFlag, altFlag;
        CFStringRef str;
        CFArrayRef list;
        int32_t width, w, value;
        double subsecond;
	UniChar ch;
        ch = CFStringGetCharacterAtIndex(format, curPos);
	if ('%' != ch) {
	    CFStringAppendCharacters(result, &ch, 1);
	    continue;
	}
	curPos++;
	if (curPos == length) {
            // trailing '%' with no garbage
            // -- discard since we don't want to encourage bare % at end
	    break;
	}
        ch = CFStringGetCharacterAtIndex(format, curPos);
	zeroFlag = false;
	spaceFlag = false;
	prefFlag = false;
        altFlag = false;
        width = 0;
        str = NULL;
	while (' ' == ch || '0' == ch || '~' == ch || '#' == ch) {
	    if (' ' == ch) spaceFlag = true;
	    if ('0' == ch) zeroFlag = true;
	    if ('~' == ch) prefFlag = true;
	    if ('#' == ch) altFlag = true;
            curPos++;
            if (curPos == length) {
                // trailing '%' with garbage
                // -- discard since we don't want to encourage garbage % at end
                break;
            }
            ch = CFStringGetCharacterAtIndex(format, curPos);
	}
        if (curPos == length) {
            break;
        }
        while ('0' <= ch && ch <= '9') {
            width = 10 * width + (ch - '0');
            curPos++;
            if (curPos == length) {
                // trailing '%' with garbage
                // -- discard since we don't want to encourage garbage % at end
                break;
            }
            ch = CFStringGetCharacterAtIndex(format, curPos);
        }
        if (curPos == length) {
            break;
        }
        switch (ch) {
        case '%':		// percent character
            str = CFRetain(CFSTR("%"));
            break;
        case 'd':		// day of month number
            str = CFStringCreateWithFormat(allocator, NULL, CFSTR("%d"), gdate.day);
            break;
        case 'f':		// fractional seconds
// to round or not to round? if we're displaying "which", say, millisecond, then no rounding;
// rounding also makes this somewhat more complicated to implement
            zeroFlag = true;
            if (0 == width) width = 3;
            subsecond = gdate.second - floor(gdate.second);
            for (w = 0; w < width; w++) {
                subsecond *= 10.0;
                value = (int32_t)floor(subsecond);
                subsecond -= (double)value;
                ch = '0' + value;
                CFStringAppendCharacters(result, &ch, 1);
            }
            continue;	// careful
        case 'h':		// hour number
            value = (altFlag ? (gdate.hour % 12) : (gdate.hour));
            if (altFlag && 0 == value) value = 12;
            str = CFStringCreateWithFormat(allocator, NULL, CFSTR("%d"), value);
            break;
        case 'j':		// day of year number
            str = CFStringCreateWithFormat(allocator, NULL, CFSTR("%d"), CFAbsoluteTimeGetDayOfYear(at, tz));
            break;
        case 'm':		// minute number
            str = CFStringCreateWithFormat(allocator, NULL, CFSTR("%d"), gdate.minute);
            break;
        case 'n':		// month of year number
            str = CFStringCreateWithFormat(allocator, NULL, CFSTR("%d"), gdate.month);
            break;
        case 'N':		// localized month name or abbreviation
	    value = gdate.month - 1;
	    if (NULL == formatOptions) {
		str = CFRetain(__CFStringMakeConstantString(altFlag ? ema[value] : emn[value]));
	    } else {
		list = CFDictionaryGetValue(formatOptions, (altFlag ? CFSTR("kCFMonthAbbreviationArray") : CFSTR("kCFMonthNameArray")));
		str = CFRetain(CFArrayGetValueAtIndex(list, value));
	    }
            break;
        case 'P':		// AM/PM designation
	    value = (gdate.hour < 12 ? 0 : 1);
	    if (NULL == formatOptions) {
		str = CFRetain(__CFStringMakeConstantString(eapa[value]));
	    } else {
		list = CFDictionaryGetValue(formatOptions, CFSTR("kCFAMPMDesignation"));
		str = CFRetain(CFArrayGetValueAtIndex(list, value));
	    }
            break;
        case 's':		// second number
            str = CFStringCreateWithFormat(allocator, NULL, CFSTR("%d"), (int32_t)floor(gdate.second));
            break;
        case 'w':		// ISO 8601 day of week number
            str = CFStringCreateWithFormat(allocator, NULL, CFSTR("%d"), CFAbsoluteTimeGetDayOfWeek(at, tz));
            break;
        case 'W':		// localized day of week name or abbreviation
	    value = CFAbsoluteTimeGetDayOfWeek(at, tz) - 1;
	    if (NULL == formatOptions) {
		str = CFRetain(__CFStringMakeConstantString(altFlag ? edowa[value] : edown[value]));
	    } else {
		list = CFDictionaryGetValue(formatOptions, (altFlag ? CFSTR("kCFDayAbbreviationArray") : CFSTR("kCFDayNameArray")));
		str = CFRetain(CFArrayGetValueAtIndex(list, value));
	    }
            break;
        case 'y':		// year number
            if (altFlag) {
           	zeroFlag = true;
                if (0 == width) width = 2;
                value = gdate.year % 100;
            } else {
                value = gdate.year;
            }
            str = CFStringCreateWithFormat(allocator, NULL, CFSTR("%d"), value);
            break;
        case 'z':		// time zone offset
            value = (NULL == tz) ? 0 : CFTimeZoneGetSecondsFromGMT(tz, at);
            str = CFStringCreateWithFormat(allocator, NULL, CFSTR("%c%02d%02d"), (value < 0 ? '-' : '+'), abs(value) / 3600, abs(value) / 60 % 60);
            break;
        case 'Z':		// time zone name or abbreviation
            if (NULL == tz) {
                str = altFlag ? CFRetain(CFSTR("GMT")) : CFRetain(CFSTR("GMT"));
            } else {
                str = altFlag ? CFTimeZoneCopyAbbreviation(tz, at) : CFRetain(CFTimeZoneGetName(tz));
            }
            break;
        default:		// discard
            continue;
        }
        if (0 == width) {
            width = (spaceFlag || zeroFlag) ? 2 : 1;
        }
        width -= CFStringGetLength(str);
        ch = (zeroFlag ? '0' : ' ');
        for (w = 0; w < width; w++) {
            CFStringAppendCharacters(result, &ch, 1);
        }
	CFStringAppend(result, str);
        CFRelease(str);
    }
    va_end(vargs);
    return result;
}



%%	percent character
        '~': no effect
        '#': no effect
%d	day of month number (1-31)
        '~': undefined
        '#': undefined
%f	fractional seconds
        '~': undefined
        '#': undefined
%h	hour number (0-23)
        '~': user preference (0-23, 0-24, 0-11, 0-12, 1-12)
        '#': 12-hour clock (1-12)
%j	day of year number (1-366)
        '~': undefined
        '#': undefined
%m	minute number (0-59)
        '~': undefined
        '#': undefined
%n	month of year number (1-12)
        '~': undefined
        '#': undefined
%N	localized month name
        '~': user preference: name or abbreviation
        '#': month abbreviation
%P	AM/PM designation
        '~': undefined
        '#': undefined
%s	second number (0-59)
        '~': undefined
        '#': undefined
%w	day of week number (1-7)
        '~': undefined
        '#': undefined
%W	localized day of week name
        '~': user preference: name or abbreviation
        '#': day of week abbreviation
%y	year number
        '~': user preference: 2 digit or full
        '#': two-digit year
%z	time zone offset
        '~': undefined
        '#': undefined
%Z	time zone name
        '~': user preference: name or abbreviation
        '#': time zone abbreviation



. NULL formatOptions is allowed, in which case the system language context info is used (English)
. NULL tz is allowed and means GMT as with other API
. incomplete format specs at end of format string are discarded
. bogus format specs are discarded
. if gdate is not valid, returns NULL
. the format indicator '%' is followed by optional flags characters followed by optional width specifier followed by format specifier
. there are 4 flags: ' ', '0', '~', '#'; which can be specified in any order and any number of times
. the ' ' and '0' specify padding on the left to fill out the given (or default) width
. if both ' ' and '0' are specified, the '0' takes precedence
. ' ' is the default if neither pad flag is specified and the width (default or otherwise) is greater than the length of the text
. a '0' before the width specifier is taken as a '0' flag, not part of the width
. the '#' flag specifies the alternate form of the format specifier
. the '~' flag specifies the user-preference driven form of the format specifier
. if both '#' and '~' are given, the '~' flag takes precedence
. the width specifier is a positive integer, which specifies the width of the resulting substitution in the format string for the format spec
. if no width is given, a width of 1 is assumed, unless a padding flag is given, in which case the default is 2
. the default width exceptions are: 2 for the alternate form of %y, and 3 for %f
. the '0' flag is assumed and mandatory for the alternate form of %y, and %f
. in the case of the %f format spec, the width determines the precision of the fractional seconds

. no localization reordering
. should '~' flag override all other flags?
. is there an alternate form that would be useful for %P?


%{...}	printf escape sequence?
	put printf format spec inside braces; %@ is supported; %n and '*' flag not supported
	this would be the only reason for having this be a varargs API

#endif 0

